from abc import ABC
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd

from evidently.metrics import ColumnQuantileMetric
from evidently.metrics import ColumnSummaryMetric
from evidently.metrics import ColumnValueListMetric
from evidently.metrics import ColumnValueRangeMetric
from evidently.metrics import ConflictPredictionMetric
from evidently.metrics import ConflictTargetMetric
from evidently.metrics import DatasetCorrelationsMetric
from evidently.metrics.data_integrity.column_summary_metric import NumericCharacteristics
from evidently.metrics.data_integrity.column_summary_metric import TextCharacteristics
from evidently.renderers.base_renderer import TestHtmlInfo
from evidently.renderers.base_renderer import TestRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.render_utils import get_distribution_plot_figure
from evidently.renderers.render_utils import plot_distr
from evidently.tests.base_test import BaseCheckValueTest
from evidently.tests.base_test import GroupData
from evidently.tests.base_test import GroupingTypes
from evidently.tests.base_test import Test
from evidently.tests.base_test import TestResult
from evidently.tests.base_test import TestValueCondition
from evidently.tests.utils import approx
from evidently.tests.utils import plot_check
from evidently.tests.utils import plot_correlations
from evidently.tests.utils import plot_metric_value
from evidently.tests.utils import plot_value_counts_tables
from evidently.tests.utils import plot_value_counts_tables_ref_curr
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.generators import BaseGenerator
from evidently.utils.types import Numeric

DATA_QUALITY_GROUP = GroupData("data_quality", "Data Quality", "")
GroupingTypes.TestGroup.add_value(DATA_QUALITY_GROUP)


class BaseDataQualityMetricsValueTest(BaseCheckValueTest, ABC):
    group = DATA_QUALITY_GROUP.id
    metric: ColumnSummaryMetric

    def __init__(
        self,
        column_name: str,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = ColumnSummaryMetric(column_name)


class TestConflictTarget(Test):
    group = DATA_QUALITY_GROUP.id
    name = "Test number of conflicts in target"
    metric: ConflictTargetMetric

    def __init__(self):
        self.metric = ConflictTargetMetric()

    def check(self):
        metric_result = self.metric.get_result()

        if metric_result.number_not_stable_target is None:
            test_result = TestResult.ERROR
            description = "No target in the dataset"

        elif metric_result.number_not_stable_target > 0:
            test_result = TestResult.FAIL
            description = f"Not stable target rows count is {metric_result.number_not_stable_target}"

        else:
            test_result = TestResult.SUCCESS
            description = "Target is stable"

        return TestResult(name=self.name, description=description, status=test_result)


class TestConflictPrediction(Test):
    group = DATA_QUALITY_GROUP.id
    name = "Test number of conflicts in prediction"
    metric: ConflictPredictionMetric

    def __init__(self):
        self.metric = ConflictPredictionMetric()

    def check(self):
        metric_result = self.metric.get_result()

        if metric_result.number_not_stable_prediction is None:
            test_result = TestResult.ERROR
            description = "No prediction in the dataset"

        elif metric_result.number_not_stable_prediction > 0:
            test_result = TestResult.FAIL
            description = f"Not stable prediction rows count is {metric_result.number_not_stable_prediction}"

        else:
            test_result = TestResult.SUCCESS
            description = "Prediction is stable"

        return TestResult(name=self.name, description=description, status=test_result)


class BaseDataQualityCorrelationsMetricsValueTest(BaseCheckValueTest, ABC):
    group = DATA_QUALITY_GROUP.id
    metric: DatasetCorrelationsMetric
    method: str

    def __init__(
        self,
        method: str = "pearson",
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        self.method = method
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = DatasetCorrelationsMetric()


class TestTargetPredictionCorrelation(BaseDataQualityCorrelationsMetricsValueTest):
    name = "Correlation between Target and Prediction"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference = self.metric.get_result().reference

        if reference is not None:
            value = reference.stats[self.method].target_prediction_correlation

            if value is not None:
                return TestValueCondition(eq=approx(value, absolute=0.25))

        return TestValueCondition(gt=0)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        return self.metric.get_result().current.stats[self.method].target_prediction_correlation

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "No target or prediction in the dataset."
        return (
            f"The correlation between the target and prediction is {value:.3}. "
            f"The test threshold is {self.get_condition()}."
        )


class TestHighlyCorrelatedColumns(BaseDataQualityCorrelationsMetricsValueTest):
    name = "Highly Correlated Columns"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_correlation = self.metric.get_result().reference

        if reference_correlation is not None:
            value = reference_correlation.stats[self.method].abs_max_features_correlation

            if value is not None:
                return TestValueCondition(eq=approx(value, relative=0.1))

        return TestValueCondition(lt=0.9)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        return self.metric.get_result().current.stats[self.method].abs_max_features_correlation

    def get_description(self, value: Numeric) -> str:
        return f"The maximum correlation is {value:.3g}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestHighlyCorrelatedColumns)
class TestHighlyCorrelatedColumnsRenderer(TestRenderer):
    def render_json(self, obj: TestHighlyCorrelatedColumns) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["abs_max_num_features_correlation"] = np.round(obj.value, 3)
        return base

    def render_html(self, obj: TestHighlyCorrelatedColumns) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        current_correlations = metric_result.current.correlation[obj.method]

        if metric_result.reference is not None:
            reference_correlations: Optional[pd.DataFrame] = metric_result.reference.correlation[obj.method]

        else:
            reference_correlations = None

        fig = plot_correlations(current_correlations, reference_correlations)
        info.with_details("Highly Correlated Features", plotly_figure(title="", figure=fig))
        return info


class TestTargetFeaturesCorrelations(BaseDataQualityCorrelationsMetricsValueTest):
    name = "Correlation between Target and Features"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_correlation = self.metric.get_result().reference

        if reference_correlation is not None:
            value = reference_correlation.stats[self.method].abs_max_target_features_correlation

            if value is not None:
                return TestValueCondition(eq=approx(value, relative=0.1))

        return TestValueCondition(lt=0.9)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        return self.metric.get_result().current.stats[self.method].abs_max_target_features_correlation

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "No target in the current dataset"

        else:
            return f"The maximum correlation is {value:.3g}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestTargetFeaturesCorrelations)
class TestTargetFeaturesCorrelationsRenderer(TestRenderer):
    def render_json(self, obj: TestTargetFeaturesCorrelations) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()

        if obj.value is not None:
            abs_max_target_features_correlation = np.round(obj.value, 3)

        else:
            abs_max_target_features_correlation = obj.value

        base["parameters"]["abs_max_target_features_correlation"] = abs_max_target_features_correlation
        return base

    def render_html(self, obj: TestTargetFeaturesCorrelations) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        current_correlations = metric_result.current.correlation[obj.method]

        if metric_result.reference is not None:
            reference_correlations: Optional[pd.DataFrame] = metric_result.reference.correlation[obj.method]

        else:
            reference_correlations = None

        fig = plot_correlations(current_correlations, reference_correlations)
        info.with_details("Target Features Correlations", plotly_figure(title="", figure=fig))
        return info


class TestPredictionFeaturesCorrelations(BaseDataQualityCorrelationsMetricsValueTest):
    name = "Correlation between Prediction and Features"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_correlation = self.metric.get_result().reference

        if reference_correlation is not None:
            value = reference_correlation.stats[self.method].abs_max_prediction_features_correlation

            if value is not None:
                return TestValueCondition(eq=approx(value, relative=0.1))

        return TestValueCondition(lt=0.9)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        return self.metric.get_result().current.stats[self.method].abs_max_prediction_features_correlation

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "No prediction in the current dataset"

        else:
            return f"The maximum correlation is {value:.3g}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestPredictionFeaturesCorrelations)
class TestPredictionFeaturesCorrelationsRenderer(TestRenderer):
    def render_json(self, obj: TestPredictionFeaturesCorrelations) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()

        if obj.value is not None:
            abs_max_prediction_features_correlation = np.round(obj.value, 3)

        else:
            abs_max_prediction_features_correlation = obj.value

        base["parameters"]["abs_max_prediction_features_correlation"] = abs_max_prediction_features_correlation
        return base

    def render_html(self, obj: TestTargetFeaturesCorrelations) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        current_correlations = metric_result.current.correlation[obj.method]

        if metric_result.reference is not None:
            reference_correlations: Optional[pd.DataFrame] = metric_result.reference.correlation[obj.method]

        else:
            reference_correlations = None

        fig = plot_correlations(current_correlations, reference_correlations)
        info.with_details("Target-Features Correlations", plotly_figure(title="", figure=fig))
        return info


class TestCorrelationChanges(BaseDataQualityCorrelationsMetricsValueTest):
    group = DATA_QUALITY_GROUP.id
    name = "Change in Correlation"
    metric: DatasetCorrelationsMetric
    corr_diff: float

    def __init__(
        self,
        corr_diff: float = 0.25,
        method: str = "pearson",
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(
            method=method,
            eq=eq,
            gt=gt,
            gte=gte,
            is_in=is_in,
            lt=lt,
            lte=lte,
            not_eq=not_eq,
            not_in=not_in,
        )
        self.corr_diff = corr_diff

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        metric_result = self.metric.get_result()

        if metric_result.reference is None:
            raise ValueError("Reference should be present")

        current_correlations = metric_result.current.correlation[self.method]
        reference_correlations: Optional[pd.DataFrame] = metric_result.reference.correlation[self.method]
        diff = reference_correlations - current_correlations
        return (diff.abs() > self.corr_diff).sum().sum() / 2

    def get_description(self, value: Numeric) -> str:
        return f"The number of correlation violations is {value:.3g}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestCorrelationChanges)
class TestCorrelationChangesRenderer(TestRenderer):
    def render_html(self, obj: TestCorrelationChanges) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        current_correlations = metric_result.current.correlation[obj.method]

        if metric_result.reference is not None:
            reference_correlations: Optional[pd.DataFrame] = metric_result.reference.correlation[obj.method]

        else:
            reference_correlations = None

        fig = plot_correlations(current_correlations, reference_correlations)
        info.with_details("Target-Features Correlations", plotly_figure(title="", figure=fig))
        return info


class BaseFeatureDataQualityMetricsTest(BaseDataQualityMetricsValueTest, ABC):
    column_name: str

    def __init__(
        self,
        column_name: str,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        self.column_name = column_name
        super().__init__(
            column_name=column_name,
            eq=eq,
            gt=gt,
            gte=gte,
            is_in=is_in,
            lt=lt,
            lte=lte,
            not_eq=not_eq,
            not_in=not_in,
        )

    def groups(self) -> Dict[str, str]:
        return {
            GroupingTypes.ByFeature.id: self.column_name,
        }

    def check(self):
        result = TestResult(name=self.name, description="The test was not launched", status=TestResult.SKIPPED)
        # features_stats = self.metric.get_result().features_stats.get_all_features()

        # if self.column_name not in features_stats:
        #     result.mark_as_fail(f"Feature '{self.column_name}' was not found")
        #     return result

        result = super().check()

        if self.value is None:
            result.mark_as_error(f"No value for the feature '{self.column_name}'")
            return result

        return result


class TestColumnValueMin(BaseFeatureDataQualityMetricsTest):
    name = "Min Value"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        ref_features_stats = self.metric.get_result().reference_characteristics
        if ref_features_stats is not None:
            if not isinstance(ref_features_stats, NumericCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical or bool")
            min_value = ref_features_stats.min
            return TestValueCondition(gte=min_value)
        raise ValueError("Neither required test parameters nor reference data has been provided.")

    def calculate_value_for_test(self) -> Optional[Union[Numeric, bool]]:
        features_stats = self.metric.get_result().current_characteristics
        if not isinstance(features_stats, NumericCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical or bool")
        min_value = features_stats.min
        return min_value

    def get_description(self, value: Numeric) -> str:
        return f"The minimum value of the column **{self.column_name}** is {value} The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestColumnValueMin)
class TestColumnValueMinRenderer(TestRenderer):
    def render_html(self, obj: TestColumnValueMin) -> TestHtmlInfo:
        column_name = obj.column_name
        info = super().render_html(obj)
        bins_for_hist = obj.metric.get_result().plot_data.bins_for_hist
        if bins_for_hist is None:
            raise ValueError(f"{column_name} should be numerical or bool")
        curr_distr = bins_for_hist["current"]
        ref_distr = None
        if "reference" in bins_for_hist.keys():
            ref_distr = bins_for_hist["reference"]
        fig = plot_distr(hist_curr=curr_distr, hist_ref=ref_distr, color_options=self.color_options)
        fig = plot_check(fig, obj.get_condition(), color_options=self.color_options)
        current_characteristics = obj.metric.get_result().current_characteristics
        if not isinstance(current_characteristics, NumericCharacteristics):
            raise ValueError(f"{column_name} should be numerical or bool")
        min_value = current_characteristics.min

        if min_value is not None:
            fig = plot_metric_value(fig, float(min_value), f"current {column_name} min value")
        info.with_details(f"Min Value {column_name}", plotly_figure(title="", figure=fig))
        return info


class TestColumnValueMax(BaseFeatureDataQualityMetricsTest):
    name = "Max Value"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        ref_features_stats = self.metric.get_result().reference_characteristics
        if ref_features_stats is not None:
            if not isinstance(ref_features_stats, NumericCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical or bool")
            max_value = ref_features_stats.max
            return TestValueCondition(lte=max_value)
        raise ValueError("Neither required test parameters nor reference data has been provided.")

    def calculate_value_for_test(self) -> Optional[Union[Numeric, bool]]:
        features_stats = self.metric.get_result().current_characteristics
        if not isinstance(features_stats, NumericCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical or bool")
        max_value = features_stats.max
        if isinstance(max_value, str):
            raise ValueError(f"{self.column_name} should be numerical or bool")
        return max_value

    def get_description(self, value: Numeric) -> str:
        return f"The maximum value of the column **{self.column_name}** is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestColumnValueMax)
class TestColumnValueMaxRenderer(TestRenderer):
    def render_html(self, obj: TestColumnValueMax) -> TestHtmlInfo:
        column_name = obj.column_name
        info = super().render_html(obj)
        bins_for_hist = obj.metric.get_result().plot_data.bins_for_hist
        if bins_for_hist is None:
            raise ValueError(f"{column_name} should be numerical or bool")
        curr_distr = bins_for_hist["current"]
        ref_distr = None
        if "reference" in bins_for_hist.keys():
            ref_distr = bins_for_hist["reference"]
        fig = plot_distr(hist_curr=curr_distr, hist_ref=ref_distr, color_options=self.color_options)
        fig = plot_check(fig, obj.get_condition(), color_options=self.color_options)
        current_characteristics = obj.metric.get_result().current_characteristics
        if not isinstance(current_characteristics, NumericCharacteristics):
            raise ValueError(f"{column_name} should be numerical or bool")
        max_value = current_characteristics.max

        if max_value is not None:
            fig = plot_metric_value(fig, float(max_value), f"current {column_name} max value")
        info.with_details(f"Max Value {column_name}", plotly_figure(title="", figure=fig))
        return info


class TestColumnValueMean(BaseFeatureDataQualityMetricsTest):
    name = "Mean Value"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        ref_features_stats = self.metric.get_result().reference_characteristics
        if ref_features_stats is not None:
            if not isinstance(ref_features_stats, NumericCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical or bool")
            return TestValueCondition(eq=approx(ref_features_stats.mean, 0.1))
        raise ValueError("Neither required test parameters nor reference data has been provided.")

    def calculate_value_for_test(self) -> Optional[Numeric]:
        features_stats = self.metric.get_result().current_characteristics
        if not isinstance(features_stats, NumericCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical or bool")
        return features_stats.mean

    def get_description(self, value: Numeric) -> str:
        return f"The mean value of the column **{self.column_name}** is {value:.3g}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestColumnValueMean)
class TestColumnValueMeanRenderer(TestRenderer):
    def render_html(self, obj: TestColumnValueMean) -> TestHtmlInfo:
        column_name = obj.column_name
        info = super().render_html(obj)
        bins_for_hist = obj.metric.get_result().plot_data.bins_for_hist
        if bins_for_hist is None:
            raise ValueError(f"{column_name} should be numerical or bool")
        curr_distr = bins_for_hist["current"]
        ref_distr = None
        if "reference" in bins_for_hist.keys():
            ref_distr = bins_for_hist["reference"]
        fig = plot_distr(hist_curr=curr_distr, hist_ref=ref_distr, color_options=self.color_options)
        fig = plot_check(fig, obj.get_condition(), color_options=self.color_options)
        current_characteristics = obj.metric.get_result().current_characteristics
        if not isinstance(current_characteristics, NumericCharacteristics):
            raise ValueError(f"{column_name} should be numerical or bool")
        mean_value = current_characteristics.mean

        if mean_value is not None:
            fig = plot_metric_value(fig, mean_value, f"current {column_name} mean value")
        info.with_details(f"Mean Value {column_name}", plotly_figure(title="", figure=fig))
        return info


class TestColumnValueMedian(BaseFeatureDataQualityMetricsTest):
    name = "Median Value"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        ref_features_stats = self.metric.get_result().reference_characteristics
        if ref_features_stats is not None:
            if not isinstance(ref_features_stats, NumericCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical or bool")
            return TestValueCondition(eq=approx(ref_features_stats.p50, 0.1))
        raise ValueError("Neither required test parameters nor reference data has been provided.")

    def calculate_value_for_test(self) -> Optional[Numeric]:
        features_stats = self.metric.get_result().current_characteristics
        if not isinstance(features_stats, NumericCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical or bool")
        return features_stats.p50

    def get_description(self, value: Numeric) -> str:
        return f"The median value of the column **{self.column_name}** is {value:.3g}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestColumnValueMedian)
class TestColumnValueMedianRenderer(TestRenderer):
    def render_html(self, obj: TestColumnValueMedian) -> TestHtmlInfo:
        column_name = obj.column_name
        info = super().render_html(obj)
        bins_for_hist = obj.metric.get_result().plot_data.bins_for_hist
        if bins_for_hist is None:
            raise ValueError(f"{column_name} should be numerical or bool")
        curr_distr = bins_for_hist["current"]
        ref_distr = None
        if "reference" in bins_for_hist.keys():
            ref_distr = bins_for_hist["reference"]
        fig = plot_distr(hist_curr=curr_distr, hist_ref=ref_distr, color_options=self.color_options)
        fig = plot_check(fig, obj.get_condition(), color_options=self.color_options)
        current_characteristics = obj.metric.get_result().current_characteristics
        if not isinstance(current_characteristics, NumericCharacteristics):
            raise ValueError(f"{column_name} should be numerical or bool")
        percentile_50 = current_characteristics.p50

        if percentile_50 is not None:
            fig = plot_metric_value(fig, percentile_50, f"current {column_name} median value")
        info.with_details(f"Median Value {column_name}", plotly_figure(title="", figure=fig))
        return info


class TestColumnValueStd(BaseFeatureDataQualityMetricsTest):
    name = "Standard Deviation (SD)"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        ref_features_stats = self.metric.get_result().reference_characteristics
        if ref_features_stats is not None:
            if not isinstance(ref_features_stats, NumericCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical or bool")
            return TestValueCondition(eq=approx(ref_features_stats.std, 0.1))
        raise ValueError("Neither required test parameters nor reference data has been provided.")

    def calculate_value_for_test(self) -> Optional[Numeric]:
        features_stats = self.metric.get_result().current_characteristics
        if not isinstance(features_stats, NumericCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical or bool")
        return features_stats.std

    def get_description(self, value: Numeric) -> str:
        return (
            f"The standard deviation of the column **{self.column_name}** is {value:.3g}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestColumnValueStd)
class TestColumnValueStdRenderer(TestRenderer):
    def render_html(self, obj: TestColumnValueStd) -> TestHtmlInfo:
        column_name = obj.column_name
        info = super().render_html(obj)
        bins_for_hist = obj.metric.get_result().plot_data.bins_for_hist
        if bins_for_hist is None:
            raise ValueError(f"{column_name} should be numerical or bool")
        curr_distr = bins_for_hist["current"]
        ref_distr = None
        if "reference" in bins_for_hist.keys():
            ref_distr = bins_for_hist["reference"]
        fig = plot_distr(hist_curr=curr_distr, hist_ref=ref_distr, color_options=self.color_options)
        info.with_details(f"Std Value {column_name}", plotly_figure(title="", figure=fig))
        return info


class TestNumberOfUniqueValues(BaseFeatureDataQualityMetricsTest):
    name = "Number of Unique Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_features_stats = self.metric.get_result().reference_characteristics
        if reference_features_stats is not None:
            if isinstance(reference_features_stats, TextCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical, categorical or datetime")
            unique_count = reference_features_stats.unique
            return TestValueCondition(eq=approx(unique_count, relative=0.1))

        return TestValueCondition(gt=1)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        features_stats = self.metric.get_result().current_characteristics
        if isinstance(features_stats, TextCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical, categorical or datetime")
        return features_stats.unique

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of the unique values in the column **{self.column_name}** is {value}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestNumberOfUniqueValues)
class TestNumberOfUniqueValuesRenderer(TestRenderer):
    def render_html(self, obj: TestNumberOfUniqueValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        column_name = obj.column_name
        counts_data = obj.metric.get_result().plot_data.counts_of_values
        if counts_data is not None:
            curr_df = counts_data["current"]
            ref_df = None
            if "reference" in counts_data.keys():
                ref_df = counts_data["reference"]
            info.details = plot_value_counts_tables_ref_curr(column_name, curr_df, ref_df, "num_of_unique_vals")

        return info


class TestUniqueValuesShare(BaseFeatureDataQualityMetricsTest):
    name = "Share of Unique Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_features_stats = self.metric.get_result().reference_characteristics
        if reference_features_stats is not None:
            if isinstance(reference_features_stats, TextCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical, categorical or datetime")
            unique_percentage = reference_features_stats.unique_percentage

            if unique_percentage is not None:
                return TestValueCondition(eq=approx(unique_percentage / 100.0, relative=0.1))

        raise ValueError("Neither required test parameters nor reference data has been provided.")

    def calculate_value_for_test(self) -> Optional[Numeric]:
        features_stats = self.metric.get_result().current_characteristics
        if isinstance(features_stats, TextCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical, categorical or datetime")
        unique_percentage = features_stats.unique_percentage

        if unique_percentage is None:
            return None

        return unique_percentage / 100.0

    def get_description(self, value: Numeric) -> str:
        return (
            f"The share of the unique values in the column **{self.column_name}** is {value:.3}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestUniqueValuesShare)
class TestUniqueValuesShareRenderer(TestRenderer):
    def render_html(self, obj: TestUniqueValuesShare) -> TestHtmlInfo:
        info = super().render_html(obj)
        column_name = obj.column_name
        counts_data = obj.metric.get_result().plot_data.counts_of_values
        if counts_data is not None:
            curr_df = counts_data["current"]
            ref_df = None
            if "reference" in counts_data.keys():
                ref_df = counts_data["reference"]
            info.details = plot_value_counts_tables_ref_curr(column_name, curr_df, ref_df, "unique_vals_sare")

        return info


class TestMostCommonValueShare(BaseFeatureDataQualityMetricsTest):
    name = "Share of the Most Common Value"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_features_stats = self.metric.get_result().reference_characteristics
        if reference_features_stats is not None:
            if isinstance(reference_features_stats, TextCharacteristics):
                raise ValueError(f"{self.column_name} should be numerical, categorical or datetime")
            most_common_percentage = reference_features_stats.most_common_percentage

            if most_common_percentage is not None:
                return TestValueCondition(eq=approx(most_common_percentage / 100.0, relative=0.1))

        return TestValueCondition(lt=0.8)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        features_stats = self.metric.get_result().current_characteristics
        if isinstance(features_stats, TextCharacteristics):
            raise ValueError(f"{self.column_name} should be numerical, categorical or datetime")
        most_common_percentage = features_stats.most_common_percentage

        if most_common_percentage is None:
            return None

        return most_common_percentage / 100.0

    def get_description(self, value: Numeric) -> str:
        counts_data = self.metric.get_result().plot_data.counts_of_values
        if counts_data is None:
            raise ValueError("counts_of_values should be provided")
        most_common_value = counts_data["current"].iloc[0, 0]
        return (
            f"The most common value in the column **{self.column_name}** is {most_common_value}. "
            f"Its share is {value:.3g}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestMostCommonValueShare)
class TestMostCommonValueShareRenderer(TestRenderer):
    def render_json(self, obj: TestMostCommonValueShare) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["column_name"] = obj.column_name
        base["parameters"]["share_most_common_value"] = obj.value
        return base

    def render_html(self, obj: TestMostCommonValueShare) -> TestHtmlInfo:
        info = super().render_html(obj)
        column_name = obj.column_name

        if column_name is None:
            raise ValueError("column_name should be present")

        counts_data = obj.metric.get_result().plot_data.counts_of_values
        if counts_data is not None:
            curr_df = counts_data["current"]
            ref_df = None
            if "reference" in counts_data.keys():
                ref_df = counts_data["reference"]
            additional_plots = plot_value_counts_tables_ref_curr(column_name, curr_df, ref_df, "most_common_value_sare")
            info.details = additional_plots
        return info


class TestAllColumnsMostCommonValueShare(BaseGenerator):
    """Creates most common value share tests for each column in the dataset"""

    columns: Optional[List[str]]

    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns

    def generate(self, columns_info: DatasetColumns) -> List[TestMostCommonValueShare]:
        if self.columns is None:
            columns = columns_info.get_all_columns_list(skip_text_columns=True)

        else:
            columns = self.columns

        return [TestMostCommonValueShare(column_name=name) for name in columns]


class TestMeanInNSigmas(Test):
    group = DATA_QUALITY_GROUP.id
    name = "Mean Value Stability"
    metric: ColumnSummaryMetric
    column_name: str
    n_sigmas: int

    def __init__(self, column_name: str, n_sigmas: int = 2):
        self.column_name = column_name
        self.n_sigmas = n_sigmas
        self.metric = ColumnSummaryMetric(column_name)

    def check(self):
        reference_feature_stats = self.metric.get_result().reference_characteristics
        features_stats = self.metric.get_result().current_characteristics

        if reference_feature_stats is None:
            test_result = TestResult.ERROR
            description = "Reference should be present"

        else:
            current_mean = features_stats.mean
            reference_mean = reference_feature_stats.mean
            reference_std = reference_feature_stats.std
            sigmas_value = reference_std * self.n_sigmas
            left_condition = reference_mean - sigmas_value
            right_condition = reference_mean + sigmas_value

            if left_condition < current_mean < right_condition:
                description = (
                    f"The mean value of the column **{self.column_name}** is {current_mean:.3g}."
                    f" The expected range is from {left_condition:.3g} to {right_condition:.3g}"
                )
                test_result = TestResult.SUCCESS

            else:
                description = (
                    f"The mean value of the column **{self.column_name}** is {current_mean:.3g}."
                    f" The expected range is from {left_condition:.3g} to {right_condition:.3g}"
                )
                test_result = TestResult.FAIL

        return TestResult(
            name=self.name,
            description=description,
            status=test_result,
            groups={GroupingTypes.ByFeature.id: self.column_name},
        )


@default_renderer(wrap_type=TestMeanInNSigmas)
class TestMeanInNSigmasRenderer(TestRenderer):
    def render_json(self, obj: TestMeanInNSigmas) -> dict:
        base = super().render_json(obj)
        metric_result = obj.metric.get_result()
        base["parameters"]["column_name"] = obj.column_name
        base["parameters"]["n_sigmas"] = obj.n_sigmas
        if not isinstance(metric_result.current_characteristics, NumericCharacteristics):
            raise ValueError(f"{obj.column_name} should be numerical or bool")
        base["parameters"]["current_mean"] = metric_result.current_characteristics.mean

        if metric_result.reference_characteristics is not None:
            if not isinstance(metric_result.reference_characteristics, NumericCharacteristics):
                raise ValueError(f"{obj.column_name} should be numerical or bool")
            base["parameters"]["reference_mean"] = metric_result.reference_characteristics.mean
            base["parameters"]["reference_std"] = metric_result.reference_characteristics.std
        return base

    def render_html(self, obj: TestMeanInNSigmas) -> TestHtmlInfo:
        column_name = obj.column_name
        metric_result = obj.metric.get_result()
        info = super().render_html(obj)

        if metric_result.reference_characteristics is None or metric_result.plot_data.bins_for_hist is None:
            return info

        if not isinstance(metric_result.reference_characteristics, NumericCharacteristics):
            raise ValueError(f"{column_name} should be numerical or bool")
        ref_mean = metric_result.reference_characteristics.mean
        ref_std = metric_result.reference_characteristics.std

        if ref_std is None or ref_mean is None:
            raise ValueError("No mean or std for reference")

        gt = ref_mean - obj.n_sigmas * ref_std
        lt = ref_mean + obj.n_sigmas * ref_std
        ref_condition = TestValueCondition(gt=gt, lt=lt)
        curr_distr = metric_result.plot_data.bins_for_hist["current"]
        ref_distr = None

        if "reference" in metric_result.plot_data.bins_for_hist.keys():
            ref_distr = metric_result.plot_data.bins_for_hist["reference"]

        fig = plot_distr(hist_curr=curr_distr, hist_ref=ref_distr, color_options=self.color_options)
        fig = plot_check(fig, ref_condition, color_options=self.color_options)
        if not isinstance(metric_result.current_characteristics, NumericCharacteristics):
            raise ValueError(f"{obj.column_name} should be numerical or bool")
        mean_value = metric_result.current_characteristics.mean

        if mean_value is not None:
            fig = plot_metric_value(fig, mean_value, f"current {column_name} mean value")

        info.with_details("", plotly_figure(title="", figure=fig))
        return info


class TestNumColumnsMeanInNSigmas(BaseGenerator):
    """Create tests of mean for all numeric columns"""

    columns: Optional[List[str]]

    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns

    def generate(self, columns_info: DatasetColumns) -> List[TestMeanInNSigmas]:
        if self.columns is None:
            columns = columns_info.num_feature_names

        else:
            columns = [column for column in self.columns if column in columns_info.num_feature_names]

        return [TestMeanInNSigmas(column_name=name, n_sigmas=2) for name in columns]


class TestValueRange(Test):
    group = DATA_QUALITY_GROUP.id
    name = "Value Range"
    metric: ColumnValueRangeMetric
    column: str
    left: Optional[float]
    right: Optional[float]

    def __init__(
        self,
        column_name: str,
        left: Optional[float] = None,
        right: Optional[float] = None,
    ):
        self.column_name = column_name
        self.left = left
        self.right = right
        self.metric = ColumnValueRangeMetric(column_name=column_name, left=left, right=right)

    def check(self):
        number_not_in_range = self.metric.get_result().current.number_not_in_range

        if number_not_in_range > 0:
            description = f"The column **{self.column_name}** has values out of range."
            test_result = TestResult.FAIL
        else:
            description = f"All values in the column **{self.column_name}** are within range"
            test_result = TestResult.SUCCESS

        return TestResult(
            name=self.name,
            description=description,
            status=test_result,
            groups={GroupingTypes.ByFeature.id: self.column_name},
        )


@default_renderer(wrap_type=TestValueRange)
class TestValueRangeRenderer(TestRenderer):
    def render_html(self, obj: TestValueRange) -> TestHtmlInfo:
        column_name = obj.column_name
        metric_result = obj.metric.get_result()
        condition_ = TestValueCondition(gt=metric_result.left, lt=metric_result.right)
        info = super().render_html(obj)
        fig = get_distribution_plot_figure(
            current_distribution=metric_result.current_distribution,
            reference_distribution=metric_result.reference_distribution,
            color_options=self.color_options,
        )
        fig = plot_check(fig, condition_, color_options=self.color_options)
        info.with_details(f"Value Range {column_name}", plotly_figure(title="", figure=fig))
        return info


class BaseDataQualityValueRangeMetricsTest(BaseCheckValueTest, ABC):
    group = DATA_QUALITY_GROUP.id
    metric: ColumnValueRangeMetric
    column: str
    left: Optional[float]
    right: Optional[float]

    def __init__(
        self,
        column_name: str,
        left: Optional[float] = None,
        right: Optional[float] = None,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        self.column_name = column_name
        self.left = left
        self.right = right
        self.metric = ColumnValueRangeMetric(column_name=column_name, left=left, right=right)

        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)

    def groups(self) -> Dict[str, str]:
        return {GroupingTypes.ByFeature.id: self.column_name}


class TestNumberOfOutRangeValues(BaseDataQualityValueRangeMetricsTest):
    name = "Number of Out-of-Range Values "

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        return TestValueCondition(eq=approx(0))

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_not_in_range

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of values out of range in the column **{self.column_name}** is {value}. "
            f" The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestNumberOfOutRangeValues)
class TestNumberOfOutRangeValuesRenderer(TestRenderer):
    def render_html(self, obj: TestNumberOfOutRangeValues) -> TestHtmlInfo:
        column_name = obj.column_name
        metric_result = obj.metric.get_result()
        info = super().render_html(obj)
        fig = get_distribution_plot_figure(
            current_distribution=metric_result.current_distribution,
            reference_distribution=metric_result.reference_distribution,
            color_options=self.color_options,
        )
        fig = plot_check(fig, obj.condition, color_options=self.color_options)
        info.with_details(f"Number Out of Range for {column_name}", plotly_figure(title="", figure=fig))
        return info


class TestShareOfOutRangeValues(BaseDataQualityValueRangeMetricsTest):
    name = "Share of Out-of-Range Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        return TestValueCondition(eq=approx(0))

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.share_not_in_range

    def get_description(self, value: Numeric) -> str:
        current_result = self.metric.get_result().current
        return (
            f"The share of values out of range in the column **{self.column_name}** is {value:.3g} "
            f"({current_result.number_not_in_range} out of {current_result.number_of_values}). "
            f" The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestShareOfOutRangeValues)
class TestShareOfOutRangeValuesRenderer(TestRenderer):
    def render_json(self, obj: TestShareOfOutRangeValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["left"] = obj.left
        base["parameters"]["right"] = obj.right
        base["parameters"]["share_not_in_range"] = obj.value
        return base

    def render_html(self, obj: TestShareOfOutRangeValues) -> TestHtmlInfo:
        column_name = obj.column_name
        metric_result = obj.metric.get_result()
        info = super().render_html(obj)
        fig = get_distribution_plot_figure(
            current_distribution=metric_result.current_distribution,
            reference_distribution=metric_result.reference_distribution,
            color_options=self.color_options,
        )
        fig = plot_check(fig, obj.condition, color_options=self.color_options)
        info.with_details(f"Share Out of Range for {column_name}", plotly_figure(title="", figure=fig))
        return info


class TestNumColumnsOutOfRangeValues(BaseGenerator):
    """Creates share of out of range values tests for all numeric columns"""

    columns: Optional[List[str]]

    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns

    def generate(self, columns_info: DatasetColumns) -> List[TestShareOfOutRangeValues]:
        if self.columns is None:
            columns = columns_info.num_feature_names

        else:
            columns = [column for column in self.columns if column in columns_info.num_feature_names]

        return [TestShareOfOutRangeValues(column_name=name) for name in columns]


class TestValueList(Test):
    group = DATA_QUALITY_GROUP.id
    name = "Out-of-List Values"
    metric: ColumnValueListMetric
    column_name: str
    values: Optional[list]

    def __init__(self, column_name: str, values: Optional[list] = None):
        self.column_name = column_name
        self.values = values
        self.metric = ColumnValueListMetric(column_name=column_name, values=values)

    def check(self):
        metric_result = self.metric.get_result()

        if metric_result.current.number_not_in_list > 0:
            test_result = TestResult.FAIL
            description = f"The column **{self.column_name}** has values out of list."

        else:
            test_result = TestResult.SUCCESS
            description = f"All values in the column **{self.column_name}** are in the list."

        return TestResult(
            name=self.name,
            description=description,
            status=test_result,
            groups={GroupingTypes.ByFeature.id: self.column_name},
        )


@default_renderer(wrap_type=TestValueList)
class TestValueListRenderer(TestRenderer):
    def render_json(self, obj: TestValueList) -> dict:
        base = super().render_json(obj)
        base["parameters"]["column_name"] = obj.column_name
        base["parameters"]["values"] = obj.values
        base["parameters"]["number_not_in_list"] = obj.metric.get_result().current.number_not_in_list
        return base

    def render_html(self, obj: TestValueList) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        column_name = metric_result.column_name
        values = metric_result.values
        curr_df = pd.DataFrame(metric_result.current.values_in_list.items(), columns=["x", "count"])

        if metric_result.reference is not None:
            ref_df = pd.DataFrame(metric_result.reference.values_in_list.items(), columns=["x", "count"])

        else:
            ref_df = None

        additional_plots = plot_value_counts_tables(column_name, values, curr_df, ref_df, "value_list")
        info.details = additional_plots
        return info


class BaseDataQualityValueListMetricsTest(BaseCheckValueTest, ABC):
    group = DATA_QUALITY_GROUP.id
    metric: ColumnValueListMetric
    column_name: str
    values: Optional[list]

    def __init__(
        self,
        column_name: str,
        values: Optional[list] = None,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        self.column_name = column_name
        self.values = values
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = ColumnValueListMetric(column_name=column_name, values=values)

    def groups(self) -> Dict[str, str]:
        return {GroupingTypes.ByFeature.id: self.column_name}


class TestNumberOfOutListValues(BaseDataQualityValueListMetricsTest):
    name = "Number Out-of-List Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        return TestValueCondition(eq=approx(0))

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_not_in_list

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of values out of list in the column **{self.column_name}** is {value}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestNumberOfOutListValues)
class TestNumberOfOutListValuesRenderer(TestRenderer):
    def render_html(self, obj: TestNumberOfOutListValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        column_name = metric_result.column_name
        values = metric_result.values
        curr_df = pd.DataFrame(metric_result.current.values_in_list.items(), columns=["x", "count"])

        if metric_result.reference is not None:
            ref_df = pd.DataFrame(metric_result.reference.values_in_list.items(), columns=["x", "count"])

        else:
            ref_df = None

        additional_plots = plot_value_counts_tables(column_name, values, curr_df, ref_df, "number_value_list")
        info.details = additional_plots
        return info


class TestShareOfOutListValues(BaseDataQualityValueListMetricsTest):
    name = "Share of Out-of-List Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        return TestValueCondition(eq=approx(0))

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.share_not_in_list

    def get_description(self, value: Numeric) -> str:
        metric_result = self.metric.get_result()
        number_not_in_range = metric_result.current.number_not_in_list
        rows_count = metric_result.current.rows_count
        return (
            f"The share of values out of list in the column **{self.column_name}** is {value:.3g} "
            f"({number_not_in_range} out of {rows_count}). "
            f"The test threshold is {self.get_condition()}."
        )


class TestCatColumnsOutOfListValues(BaseGenerator):
    """Create share of out of list values tests for category columns"""

    columns: Optional[List[str]]

    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns

    def generate(self, columns_info: DatasetColumns) -> List[TestShareOfOutListValues]:
        if self.columns is None:
            columns = columns_info.cat_feature_names

        else:
            columns = [column for column in self.columns if column in columns_info.cat_feature_names]

        return [TestShareOfOutListValues(column_name=name) for name in columns]


class TestColumnQuantile(BaseCheckValueTest):
    group = DATA_QUALITY_GROUP.id
    name = "Quantile Value"
    metric: ColumnQuantileMetric
    column_name: str
    quantile: float

    def __init__(
        self,
        column_name: str,
        quantile: float,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        self.column_name = column_name
        self.quantile = quantile
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = ColumnQuantileMetric(column_name=column_name, quantile=quantile)

    def groups(self) -> Dict[str, str]:
        return {GroupingTypes.ByFeature.id: self.column_name}

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_value = self.metric.get_result().reference

        if reference_value is not None:
            return TestValueCondition(eq=approx(reference_value, 0.1))

        raise ValueError("Neither required test parameters nor reference data has been provided.")

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current

    def get_description(self, value: Numeric) -> str:
        return (
            f"The {self.quantile} quantile value of the column **{self.column_name}** is {value:.3g}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestColumnQuantile)
class TestColumnQuantileRenderer(TestRenderer):
    def render_html(self, obj: TestColumnQuantile) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        column_name = metric_result.column_name
        fig = get_distribution_plot_figure(
            current_distribution=metric_result.current_distribution,
            reference_distribution=metric_result.reference_distribution,
            color_options=self.color_options,
        )
        fig = plot_check(fig, obj.get_condition(), color_options=self.color_options)
        fig = plot_metric_value(
            fig, obj.metric.get_result().current, f"current {column_name} {metric_result.quantile} quantile"
        )
        info.with_details("", plotly_figure(title="", figure=fig))
        return info


@default_renderer(wrap_type=TestShareOfOutListValues)
class TestShareOfOutListValuesRenderer(TestRenderer):
    def render_json(self, obj: TestShareOfOutListValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["values"] = obj.values
        base["parameters"]["share_not_in_list"] = obj.value
        return base

    def render_html(self, obj: TestShareOfOutListValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        column_name = metric_result.column_name
        values = metric_result.values
        curr_df = pd.DataFrame(metric_result.current.values_in_list.items(), columns=["x", "count"])

        if metric_result.reference is not None:
            ref_df = pd.DataFrame(metric_result.reference.values_in_list.items(), columns=["x", "count"])

        else:
            ref_df = None

        additional_plots = plot_value_counts_tables(column_name, values, curr_df, ref_df, "share_value_list")
        info.details = additional_plots
        return info



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class ProbClassPredictionCloudWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ProbClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = ProbClassificationPerformanceAnalyzer.get_results(analyzers_results)
        utility_columns = results.columns.utility_columns

        if utility_columns.target is None or utility_columns.prediction is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )
            return None

        if self.dataset == "current":
            dataset_to_plot = (
                current_data.copy(deep=False) if current_data is not None else None
            )

        else:
            dataset_to_plot = reference_data.copy(deep=False)

        if dataset_to_plot is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires reference dataset but it is None"
                )
            return None

        dataset_to_plot.replace([np.inf, -np.inf], np.nan, inplace=True)

        # plot clouds
        graphs = []

        for label in utility_columns.prediction:
            fig = go.Figure()

            fig.add_trace(
                go.Scatter(
                    x=np.random.random(
                        dataset_to_plot[
                            dataset_to_plot[utility_columns.target] == label
                        ].shape[0]
                    ),
                    y=dataset_to_plot[dataset_to_plot[utility_columns.target] == label][
                        label
                    ],
                    mode="markers",
                    name=str(label),
                    marker=dict(size=6, color=color_options.primary_color),
                )
            )

            fig.add_trace(
                go.Scatter(
                    x=np.random.random(
                        dataset_to_plot[
                            dataset_to_plot[utility_columns.target] != label
                        ].shape[0]
                    ),
                    y=dataset_to_plot[dataset_to_plot[utility_columns.target] != label][
                        label
                    ],
                    mode="markers",
                    name="other",
                    marker=dict(size=6, color=color_options.secondary_color),
                )
            )

            fig.update_layout(
                yaxis_title="Probability",
                xaxis=dict(range=(-2, 3), showticklabels=False),
            )

            fig_json = json.loads(fig.to_json())

            graphs.append(
                {
                    "id": "tab_" + str(label),
                    "title": str(label),
                    "graph": {
                        "data": fig_json["data"],
                        "layout": fig_json["layout"],
                    },
                }
            )

        return BaseWidgetInfo(
            title=self.title,
            type="tabbed_graph",
            details="",
            alerts=[],
            alertsPosition="row",
            insights=[],
            size=1 if current_data is not None else 2,
            params={"graphs": graphs},
            additionalGraphs=[],
        )



from dataclasses import dataclass
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union


class TaskType:
    REGRESSION_TASK: str = "regression"
    CLASSIFICATION_TASK: str = "classification"


@dataclass
class ColumnMapping:
    target: Optional[str] = "target"
    prediction: Optional[Union[str, int, Union[Sequence[str], Sequence[int]]]] = "prediction"
    datetime: Optional[str] = "datetime"
    id: Optional[str] = None
    numerical_features: Optional[List[str]] = None
    categorical_features: Optional[List[str]] = None
    datetime_features: Optional[List[str]] = None
    target_names: Optional[List[str]] = None
    task: Optional[str] = None
    pos_label: Optional[Union[str, int]] = 1
    text_features: Optional[List[str]] = None

    def is_classification_task(self):
        return self.task == TaskType.CLASSIFICATION_TASK

    def is_regression_task(self):
        return self.task == TaskType.REGRESSION_TASK



"""G-test of two samples.

Name: "g_test"

Import:

    >>> from evidently.calculations.stattests import g_test

Properties:
- only for categorical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import g_test
    >>> options = DataDriftOptions(all_features_stattest=g_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="g_test")
"""
from typing import Tuple

import pandas as pd
from scipy.stats import power_divergence

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest
from evidently.calculations.stattests.utils import get_unique_not_nan_values_list_from_series


def _g_stat_test(
    reference_data: pd.Series,
    current_data: pd.Series,
    feature_type: str,
    threshold: float,
) -> Tuple[float, bool]:
    """Compute the G test between two arrays
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: level of significance
    Returns:
        p_value: calculated p_value value
        test_result: whether the drift is detected
    """
    keys = get_unique_not_nan_values_list_from_series(current_data=current_data, reference_data=reference_data)
    ref_feature_dict = {**dict.fromkeys(keys, 0), **dict(reference_data.value_counts())}
    current_feature_dict = {**dict.fromkeys(keys, 0), **dict(current_data.value_counts())}
    f_exp = [ref_feature_dict[key] for key in keys]
    f_obs = [current_feature_dict[key] for key in keys]
    p_value = power_divergence(f_obs, f_exp, lambda_="log-likelihood")[1]
    return p_value, p_value < threshold


g_test = StatTest(
    name="g_test", display_name="g_test", func=_g_stat_test, allowed_feature_types=["cat"], default_threshold=0.1
)

register_stattest(g_test)



"""Jensen-Shannon distance of two samples.

Name: "jensenshannon"

Import:

    >>> from evidently.calculations.stattests import jensenshannon_stat_test

Properties:
- only for categorical and numerical features
- returns distance

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import jensenshannon_stat_test
    >>> options = DataDriftOptions(all_features_stattest=jensenshannon_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="jensenshannon")
"""
from typing import Tuple

import pandas as pd
from scipy.spatial import distance

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest
from evidently.calculations.stattests.utils import get_binned_data


def _jensenshannon(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float, n_bins: int = 30
) -> Tuple[float, bool]:
    """Compute the Jensen-Shannon distance between two arrays
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
        n_bins: number of bins
    Returns:
        jensenshannon: calculated Jensen-Shannon distance
        test_result: whether the drift is detected
    """
    reference_percents, current_percents = get_binned_data(reference_data, current_data, feature_type, n_bins, False)
    jensenshannon_value = distance.jensenshannon(reference_percents, current_percents)
    return jensenshannon_value, jensenshannon_value >= threshold


jensenshannon_stat_test = StatTest(
    name="jensenshannon",
    display_name="Jensen-Shannon distance",
    func=_jensenshannon,
    allowed_feature_types=["cat", "num"],
    default_threshold=0.1,
)

register_stattest(jensenshannon_stat_test)



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class ProbClassPRCurveWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ProbClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = ProbClassificationPerformanceAnalyzer.get_results(analyzers_results)
        utility_columns = results.columns.utility_columns

        if utility_columns.target is None or utility_columns.prediction is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )

            return None

        if self.dataset == "reference":
            metrics = results.reference_metrics

            if metrics is None:
                raise ValueError(
                    f"Widget [{self.title}] required 'reference' results from"
                    f" {ProbClassificationPerformanceAnalyzer.__name__} but no data found"
                )

        elif self.dataset == "current":
            metrics = results.current_metrics

        else:
            raise ValueError(
                f"Widget [{self.title}] required 'current' or 'reference' dataset value"
            )

        if metrics is None:
            return None

        # plot PR-curve
        if len(utility_columns.prediction) <= 2:
            if metrics.pr_curve is None:
                raise ValueError(f"Widget [{self.title}] got no pr_curve value")

            pr_curve = metrics.pr_curve
            fig = go.Figure()
            fig.add_trace(
                go.Scatter(
                    x=pr_curve["rcl"],
                    y=pr_curve["pr"],
                    mode="lines",
                    name="PR",
                    marker=dict(
                        size=6,
                        color=color_options.primary_color,
                    ),
                )
            )

            fig.update_layout(
                yaxis_title="Precision", xaxis_title="Recall", showlegend=True
            )

            fig_json = json.loads(fig.to_json())

            widget_info = BaseWidgetInfo(
                title=self.title,
                type="big_graph",
                size=1 if current_data is not None else 2,
                params={"data": fig_json["data"], "layout": fig_json["layout"]},
            )
        else:
            if metrics.pr_curve is None:
                raise ValueError(f"Widget [{self.title}] got no pr_curve value")

            if not isinstance(metrics.pr_curve, dict):
                raise ValueError(
                    f"Widget [{self.title}] got incorrect type for pr_curve value"
                )

            graphs = []

            for label in utility_columns.prediction:
                pr_curve = metrics.pr_curve[label]
                fig = go.Figure()
                fig.add_trace(
                    go.Scatter(
                        x=pr_curve["pr"],
                        y=pr_curve["rcl"],
                        mode="lines",
                        name="PR",
                        marker=dict(
                            size=6,
                            color=color_options.primary_color,
                        ),
                    )
                )

                fig.update_layout(
                    yaxis_title="Precision", xaxis_title="Recall", showlegend=True
                )

                fig_json = json.loads(fig.to_json())

                graphs.append(
                    {
                        "id": "tab_" + str(label),
                        "title": str(label),
                        "graph": {
                            "data": fig_json["data"],
                            "layout": fig_json["layout"],
                        },
                    }
                )

            widget_info = BaseWidgetInfo(
                title=self.title,
                type="tabbed_graph",
                size=1 if current_data is not None else 2,
                params={"graphs": graphs},
            )
        return widget_info



import dataclasses
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import pandas as pd
from numpy import dtype
from pandas.core.dtypes.common import is_float_dtype
from pandas.core.dtypes.common import is_integer_dtype
from pandas.core.dtypes.common import is_object_dtype
from pandas.core.dtypes.common import is_string_dtype
from sklearn import metrics

from evidently import ColumnMapping
from evidently.utils.data_operations import DatasetColumns


@dataclass
class ConfusionMatrix:
    labels: Sequence[Union[str, int]]
    values: list


def calculate_confusion_by_classes(
    confusion_matrix: np.ndarray,
    class_names: Sequence[Union[str, int]],
) -> Dict[Union[str, int], Dict[str, int]]:
    """Calculate metrics:
    - TP (true positive)
    - TN (true negative)
    - FP (false positive)
    - FN (false negative)
    for each class from confusion matrix.

    Returns:
        a dict like::

            {
                "class_1_name": {
                    "tp": 1,
                    "tn": 5,
                    "fp": 0,
                    "fn": 3,
                },
                "class_1_name": {
                    "tp": 1,
                    "tn": 5,
                    "fp": 0,
                    "fn": 3,
                },
            }
    """
    true_positive = np.diag(confusion_matrix)
    false_positive = confusion_matrix.sum(axis=0) - np.diag(confusion_matrix)
    false_negative = confusion_matrix.sum(axis=1) - np.diag(confusion_matrix)
    true_negative = confusion_matrix.sum() - (false_positive + false_negative + true_positive)
    confusion_by_classes = {}

    for idx, class_name in enumerate(class_names):
        confusion_by_classes[class_name] = {
            "tp": true_positive[idx],
            "tn": true_negative[idx],
            "fp": false_positive[idx],
            "fn": false_negative[idx],
        }

    return confusion_by_classes


def k_probability_threshold(prediction_probas: pd.DataFrame, k: Union[int, float]) -> float:
    probas = prediction_probas.iloc[:, 0].sort_values(ascending=False)
    if isinstance(k, float):
        if k < 0.0 or k > 1.0:
            raise ValueError(f"K should be in range [0.0, 1.0] but was {k}")
        return probas.iloc[max(int(np.ceil(k * prediction_probas.shape[0])) - 1, 0)]
    if isinstance(k, int):
        return probas.iloc[min(k, prediction_probas.shape[0] - 1)]
    raise ValueError(f"K has unexpected type {type(k)}")


@dataclasses.dataclass
class PredictionData:
    predictions: pd.Series
    prediction_probas: Optional[pd.DataFrame]
    labels: List[Union[str, int]]


def get_prediction_data(
    data: pd.DataFrame,
    data_columns: DatasetColumns,
    pos_label: Optional[Union[str, int]],
    threshold: float = 0.5,
) -> PredictionData:
    """Get predicted values and optional prediction probabilities from source data.
    Also take into account a threshold value - if a probability is less than the value, do not take it into account.

    Return and object with predicted values and an optional prediction probabilities.
    """
    # binary or multiclass classification
    # for binary prediction_probas has column order [pos_label, neg_label]
    # for multiclass classification return just values and probas
    prediction = data_columns.utility_columns.prediction
    target = data_columns.utility_columns.target
    if isinstance(prediction, list) and len(prediction) > 2:
        # list of columns with prediction probas, should be same as target labels
        return PredictionData(
            predictions=data[prediction].idxmax(axis=1),
            prediction_probas=data[prediction],
            labels=prediction,
        )

    # calculate labels as np.array - for better negative label calculations for binary classification
    if data_columns.target_names is not None:
        # if target_names is specified, get labels from it
        labels = np.array(data_columns.target_names)

    else:
        # if target_names is not specified, try to get labels from target and/or prediction
        if isinstance(prediction, str) and not is_float_dtype(data[prediction]):
            # if prediction is not probas, get unique values from it and target
            labels = np.union1d(data[target].unique(), data[prediction].unique())

        else:
            # if prediction is probas, get unique values from target only
            labels = data[target].unique()

    # binary classification
    # prediction in mapping is a list of two columns:
    # one is positive value probabilities, second is negative value probabilities
    if isinstance(prediction, list) and len(prediction) == 2:
        pos_label = _check_pos_labels(pos_label, labels)

        # get negative label for binary classification
        neg_label = labels[labels != pos_label][0]

        predictions = threshold_probability_labels(data[prediction], pos_label, neg_label, threshold)
        return PredictionData(
            predictions=predictions,
            prediction_probas=data[[pos_label, neg_label]],
            labels=[pos_label, neg_label],
        )

    # binary classification
    # target is strings or other values, prediction is a string with positive label name, one column with probabilities
    if (
        isinstance(prediction, str)
        and (is_string_dtype(data[target]) or is_object_dtype(data[target]))
        and is_float_dtype(data[prediction])
    ):
        pos_label = _check_pos_labels(pos_label, labels)

        if prediction not in labels:
            raise ValueError(
                "No prediction for the target labels were found. "
                "Consider to rename columns with the prediction to match target labels."
            )

        # get negative label for binary classification
        neg_label = labels[labels != pos_label][0]

        if pos_label == prediction:
            pos_preds = data[prediction]

        else:
            pos_preds = data[prediction].apply(lambda x: 1.0 - x)

        prediction_probas = pd.DataFrame.from_dict(
            {
                pos_label: pos_preds,
                neg_label: pos_preds.apply(lambda x: 1.0 - x),
            }
        )
        predictions = threshold_probability_labels(prediction_probas, pos_label, neg_label, threshold)
        return PredictionData(
            predictions=predictions,
            prediction_probas=prediction_probas,
            labels=[pos_label, neg_label],
        )

    # binary target and preds are numbers and prediction is a label
    if not isinstance(prediction, list) and prediction in [0, 1, "0", "1"] and pos_label == 0:
        if prediction in [0, "0"]:
            pos_preds = data[prediction]
        else:
            pos_preds = data[prediction].apply(lambda x: 1.0 - x)
        predictions = pos_preds.apply(lambda x: 0 if x >= threshold else 1)
        prediction_probas = pd.DataFrame.from_dict(
            {
                0: pos_preds,
                1: pos_preds.apply(lambda x: 1.0 - x),
            }
        )
        return PredictionData(
            predictions=predictions,
            prediction_probas=prediction_probas,
            labels=[0, 1],
        )

    # binary target and preds are numbers
    elif (
        isinstance(prediction, str)
        and is_integer_dtype(data[target].dtype)
        and data[prediction].dtype == dtype("float")
    ):
        predictions = (data[prediction] >= threshold).astype(dtype("int64"))
        prediction_probas = pd.DataFrame.from_dict(
            {
                1: data[prediction],
                0: data[prediction].apply(lambda x: 1.0 - x),
            }
        )
        return PredictionData(
            predictions=predictions,
            prediction_probas=prediction_probas,
            labels=[1, 0],
        )

    # for other cases return just prediction values, probabilities are None by default
    return PredictionData(
        predictions=data[prediction],
        prediction_probas=None,
        labels=data[prediction].unique().tolist(),
    )


def _check_pos_labels(pos_label: Optional[Union[str, int]], labels: List[str]) -> Union[str, int]:
    if pos_label is None:
        raise ValueError("Undefined pos_label.")

    if pos_label not in labels:
        raise ValueError(f"Cannot find pos_label '{pos_label}' in labels {labels}")

    return pos_label


def threshold_probability_labels(
    prediction_probas: pd.DataFrame, pos_label: Union[str, int], neg_label: Union[str, int], threshold: float
) -> pd.Series:
    """Get prediction values by probabilities with the threshold apply"""
    return prediction_probas[pos_label].apply(lambda x: pos_label if x >= threshold else neg_label)


STEP_SIZE = 0.05


def calculate_pr_table(binded):
    result = []
    binded.sort(key=lambda item: item[1], reverse=True)
    data_size = len(binded)
    target_class_size = sum([x[0] for x in binded])
    offset = max(round(data_size * STEP_SIZE), 1)

    for step in np.arange(offset, data_size + offset, offset):
        count = min(step, data_size)
        prob = round(binded[min(step, data_size - 1)][1], 2)
        top = round(100.0 * min(step, data_size) / data_size, 1)
        tp = sum([x[0] for x in binded[: min(step, data_size)]])
        fp = count - tp
        precision = round(100.0 * tp / count, 1)
        recall = round(100.0 * tp / target_class_size, 1)
        result.append([top, int(count), prob, int(tp), int(fp), precision, recall])
    return result


def calculate_matrix(
    target: pd.Series,
    prediction: pd.Series,
    labels: List[Union[str, int]],
) -> ConfusionMatrix:
    sorted_labels = sorted(labels)
    matrix = metrics.confusion_matrix(target, prediction, labels=sorted_labels)
    return ConfusionMatrix(sorted_labels, [row.tolist() for row in matrix])


def collect_plot_data(prediction_probas: pd.DataFrame):
    res = {}
    mins = []
    lowers = []
    means = []
    uppers = []
    maxs = []
    for col in prediction_probas.columns:
        mins.append(np.percentile(prediction_probas[col], 0))
        lowers.append(np.percentile(prediction_probas[col], 25))
        means.append(np.percentile(prediction_probas[col], 50))
        uppers.append(np.percentile(prediction_probas[col], 75))
        maxs.append(np.percentile(prediction_probas[col], 100))
    res["mins"] = mins
    res["lowers"] = lowers
    res["means"] = means
    res["uppers"] = uppers
    res["maxs"] = maxs
    return res


@dataclasses.dataclass
class DatasetClassificationQuality:
    accuracy: float
    precision: float
    recall: float
    f1: float
    roc_auc: Optional[float] = None
    log_loss: Optional[float] = None
    tpr: Optional[float] = None
    tnr: Optional[float] = None
    fpr: Optional[float] = None
    fnr: Optional[float] = None
    rate_plots_data: Optional[Dict] = None
    plot_data: Optional[Dict] = None


def calculate_metrics(
    column_mapping: ColumnMapping,
    confusion_matrix: ConfusionMatrix,
    target: pd.Series,
    prediction: PredictionData,
) -> DatasetClassificationQuality:
    if column_mapping.pos_label is not None:
        pos_label = column_mapping.pos_label
    else:
        pos_label = 1
    tpr = None
    tnr = None
    fpr = None
    fnr = None
    roc_auc = None
    log_loss = None
    rate_plots_data = None
    plot_data = None
    if len(prediction.labels) == 2:
        confusion_by_classes = calculate_confusion_by_classes(
            np.array(confusion_matrix.values),
            confusion_matrix.labels,
        )
        conf_by_pos_label = confusion_by_classes[pos_label]
        precision = metrics.precision_score(target, prediction.predictions, pos_label=pos_label)
        recall = metrics.recall_score(target, prediction.predictions, pos_label=pos_label)
        f1 = metrics.f1_score(target, prediction.predictions, pos_label=pos_label)
        tpr = conf_by_pos_label["tp"] / (conf_by_pos_label["tp"] + conf_by_pos_label["fn"])
        tnr = conf_by_pos_label["tn"] / (conf_by_pos_label["tn"] + conf_by_pos_label["fp"])
        fpr = conf_by_pos_label["fp"] / (conf_by_pos_label["fp"] + conf_by_pos_label["tn"])
        fnr = conf_by_pos_label["fn"] / (conf_by_pos_label["fn"] + conf_by_pos_label["tp"])
    else:
        precision = metrics.precision_score(target, prediction.predictions, average="macro")
        recall = metrics.recall_score(target, prediction.predictions, average="macro")
        f1 = metrics.f1_score(target, prediction.predictions, average="macro")
    if prediction.prediction_probas is not None:
        binaraized_target = (
            target.astype(str).values.reshape(-1, 1) == list(prediction.prediction_probas.columns.astype(str))
        ).astype(int)
        prediction_probas_array = prediction.prediction_probas.to_numpy()
        roc_auc = metrics.roc_auc_score(binaraized_target, prediction_probas_array, average="macro")
        log_loss = metrics.log_loss(binaraized_target, prediction_probas_array)
        plot_data = collect_plot_data(prediction.prediction_probas)
    if len(prediction.labels) == 2 and prediction.prediction_probas is not None:
        fprs, tprs, thrs = metrics.roc_curve(target == pos_label, prediction.prediction_probas[pos_label])
        roc_curve = {"fpr": fprs.tolist(), "tpr": tprs.tolist(), "thrs": thrs.tolist()}
        rate_plots_data = {
            "thrs": roc_curve["thrs"],
            "tpr": roc_curve["tpr"],
            "fpr": roc_curve["fpr"],
        }
        df = pd.DataFrame(
            {"true": (target == pos_label).astype(int).values, "preds": prediction.prediction_probas[pos_label].values}
        )
        tnrs = []
        fnrs = []
        for tr in rate_plots_data["thrs"]:
            if tr < 1:
                tn = df[(df.true == 0) & (df.preds < tr)].shape[0]
                fn = df[(df.true == 1) & (df.preds < tr)].shape[0]
                tp = df[(df.true == 1) & (df.preds >= tr)].shape[0]
                fp = df[(df.true == 0) & (df.preds >= tr)].shape[0]
                tnrs.append(tn / (tn + fp))
                fnrs.append(fn / (fn + tp))
            else:
                fnrs.append(1)
                tnrs.append(1)
        rate_plots_data["fnr"] = fnrs
        rate_plots_data["tnr"] = tnrs

    return DatasetClassificationQuality(
        accuracy=metrics.accuracy_score(target, prediction.predictions),
        precision=precision,
        recall=recall,
        f1=f1,
        tpr=tpr,
        tnr=tnr,
        fpr=fpr,
        fnr=fnr,
        roc_auc=roc_auc,
        log_loss=log_loss,
        rate_plots_data=rate_plots_data,
        plot_data=plot_data,
    )



from typing import Dict
from typing import List
from typing import Optional

from evidently import TaskType
from evidently.base_metric import InputData
from evidently.calculations.stattests import PossibleStatTestType
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestAllFeaturesValueDrift
from evidently.tests import TestColumnDrift
from evidently.tests import TestShareOfDriftedColumns
from evidently.utils.data_drift_utils import resolve_stattest_threshold
from evidently.utils.data_operations import DatasetColumns


class DataDriftTestPreset(TestPreset):
    """
    Data Drift tests.

    Contains tests:
    - `TestShareOfDriftedColumns`
    - `TestColumnValueDrift`
    - `TestAllFeaturesValueDrift`
    """

    columns: Optional[List[str]]
    drift_share: Optional[float]
    stattest: Optional[PossibleStatTestType]
    cat_stattest: Optional[PossibleStatTestType]
    num_stattest: Optional[PossibleStatTestType]
    text_stattest: Optional[PossibleStatTestType]
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]]
    stattest_threshold: Optional[float]
    cat_stattest_threshold: Optional[float]
    num_stattest_threshold: Optional[float]
    text_stattest_threshold: Optional[float]
    per_column_stattest_threshold: Optional[Dict[str, float]]

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        drift_share: Optional[float] = None,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        super().__init__()
        self.columns = columns
        self.drift_share = drift_share
        self.stattest = stattest
        self.cat_stattest = cat_stattest
        self.num_stattest = num_stattest
        self.text_stattest = text_stattest
        self.per_column_stattest = per_column_stattest
        self.stattest_threshold = stattest_threshold
        self.cat_stattest_threshold = cat_stattest_threshold
        self.num_stattest_threshold = num_stattest_threshold
        self.text_stattest_threshold = text_stattest_threshold
        self.per_column_stattest_threshold = per_column_stattest_threshold

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        preset_tests: list = [
            TestShareOfDriftedColumns(
                columns=self.columns,
                lt=0.3 if self.drift_share is None else self.drift_share,
                stattest=self.stattest,
                cat_stattest=self.cat_stattest,
                num_stattest=self.num_stattest,
                text_stattest=self.text_stattest,
                per_column_stattest=self.per_column_stattest,
                stattest_threshold=self.stattest_threshold,
                cat_stattest_threshold=self.cat_stattest_threshold,
                num_stattest_threshold=self.num_stattest_threshold,
                text_stattest_threshold=self.text_stattest_threshold,
                per_column_stattest_threshold=self.per_column_stattest_threshold,
            ),
        ]

        if columns.utility_columns.target is not None:
            stattest, threshold = resolve_stattest_threshold(
                columns.utility_columns.target,
                "cat" if columns.task == TaskType.CLASSIFICATION_TASK else "num",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            preset_tests.append(
                TestColumnDrift(
                    column_name=columns.utility_columns.target,
                    stattest_threshold=threshold,
                    stattest=stattest,
                )
            )

        if columns.utility_columns.prediction is not None and isinstance(columns.utility_columns.prediction, str):
            stattest, threshold = resolve_stattest_threshold(
                columns.utility_columns.prediction,
                "cat" if columns.task == TaskType.CLASSIFICATION_TASK else "num",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            preset_tests.append(
                TestColumnDrift(
                    column_name=columns.utility_columns.prediction,
                    stattest_threshold=threshold,
                    stattest=stattest,
                )
            )

        preset_tests.append(
            TestAllFeaturesValueDrift(
                self.columns,
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
        )

        return preset_tests



#!/usr/bin/env python
# coding: utf-8


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "nbextension/static",
            "dest": "evidently",
            "require": "evidently/extension",
        }
    ]



#!/usr/bin/env python
# coding: utf-8
import copy
from dataclasses import dataclass
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.base_analyzer import BaseAnalyzerResult
from evidently.calculations.regression_performance import RegressionPerformanceMetrics
from evidently.calculations.regression_performance import calculate_regression_performance
from evidently.utils.data_operations import process_columns


@dataclass
class RegressionPerformanceAnalyzerResults(BaseAnalyzerResult):
    reference_metrics: Optional[RegressionPerformanceMetrics] = None
    current_metrics: Optional[RegressionPerformanceMetrics] = None
    error_bias: Optional[dict] = None


class RegressionPerformanceAnalyzer(Analyzer):
    @staticmethod
    def get_results(analyzer_results) -> RegressionPerformanceAnalyzerResults:
        return analyzer_results[RegressionPerformanceAnalyzer]

    def calculate(
        self, reference_data: pd.DataFrame, current_data: Optional[pd.DataFrame], column_mapping: ColumnMapping
    ) -> RegressionPerformanceAnalyzerResults:
        columns = process_columns(reference_data, column_mapping)
        result = RegressionPerformanceAnalyzerResults(columns=columns)

        target_column = columns.utility_columns.target
        prediction_column = columns.utility_columns.prediction

        if target_column is not None and prediction_column is not None:
            result.reference_metrics = calculate_regression_performance(
                dataset=reference_data, columns=columns, error_bias_prefix="ref_"
            )

            if result.reference_metrics.error_bias is not None:
                error_bias = copy.deepcopy(result.reference_metrics.error_bias)

            else:
                error_bias = None

            if current_data is not None:
                result.current_metrics = calculate_regression_performance(
                    dataset=current_data, columns=columns, error_bias_prefix="current_"
                )
                if result.current_metrics.error_bias is not None:
                    if error_bias is not None:
                        for feature_name, current_bias in result.current_metrics.error_bias.items():
                            error_bias[feature_name].update(current_bias)

                    else:
                        error_bias = copy.deepcopy(result.current_metrics.error_bias)

            if error_bias:
                result.error_bias = error_bias

            else:
                result.error_bias = {}

        return result



from typing import Optional

from evidently.base_metric import InputData
from evidently.calculations.stattests import PossibleStatTestType
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestAccuracyScore
from evidently.tests import TestColumnDrift
from evidently.tests import TestF1Score
from evidently.tests import TestPrecisionScore
from evidently.tests import TestRecallScore
from evidently.tests import TestRocAuc
from evidently.utils.data_operations import DatasetColumns


class BinaryClassificationTestPreset(TestPreset):
    """
    Binary Classification Tests.
    Args:
        stattest: statistical test for target drift test.
        stattest_threshold: threshold for statistical test for target drift test.
        probas_threshold: threshold for label calculation for prediction.

    Contains:
    - `TestColumnValueDrift` for target
    - `TestPrecisionScore`
    - `TestRecallScore`
    - `TestF1Score`
    - `TestAccuracyScore`
    """

    def __init__(
        self,
        stattest: Optional[PossibleStatTestType] = None,
        stattest_threshold: Optional[float] = None,
        probas_threshold: Optional[float] = None,
    ):
        super().__init__()
        self.stattest = stattest
        self.stattest_threshold = stattest_threshold
        self.probas_threshold = probas_threshold

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        target = columns.utility_columns.target

        if target is None:
            raise ValueError("Target column should be set in mapping and be present in data")
        prediction_columns = data.data_definition.get_prediction_columns()
        is_probas_present = prediction_columns is not None and prediction_columns.prediction_probas is not None
        if not is_probas_present:
            return [
                TestColumnDrift(column_name=target, stattest=self.stattest, stattest_threshold=self.stattest_threshold),
                TestPrecisionScore(probas_threshold=self.probas_threshold),
                TestRecallScore(probas_threshold=self.probas_threshold),
                TestF1Score(probas_threshold=self.probas_threshold),
                TestAccuracyScore(probas_threshold=self.probas_threshold),
            ]

        return [
            TestColumnDrift(column_name=target, stattest=self.stattest, stattest_threshold=self.stattest_threshold),
            TestRocAuc(),
            TestPrecisionScore(probas_threshold=self.probas_threshold),
            TestRecallScore(probas_threshold=self.probas_threshold),
            TestAccuracyScore(probas_threshold=self.probas_threshold),
            TestF1Score(probas_threshold=self.probas_threshold),
        ]



#!/usr/bin/env python
# coding: utf-8
from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.data_drift_table_widget import DataDriftTableWidget


class DataDriftTab(Tab):
    widgets = [(DataDriftTableWidget("Data Drift"), Verbose.ALWAYS)]



from typing import Dict
from typing import Optional
from typing import Type
from typing import Union

from evidently.base_metric import Metric
from evidently.utils.generators import BaseGenerator
from evidently.utils.generators import make_generator_by_columns


def generate_column_metrics(
    metric_class: Type[Metric],
    columns: Optional[Union[str, list]] = None,
    parameters: Optional[Dict] = None,
    skip_id_column: bool = False,
) -> BaseGenerator:
    """Function for generating metrics for columns"""
    return make_generator_by_columns(
        base_class=metric_class,
        columns=columns,
        parameters=parameters,
        skip_id_column=skip_id_column,
    )



import itertools
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Type

import pandas

from evidently.analyzers.base_analyzer import Analyzer
from evidently.options import OptionsProvider
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.pipeline.stage import PipelineStage


class Pipeline:
    _analyzers: List[Type[Analyzer]]
    stages: Sequence[PipelineStage]
    analyzers_results: Dict[Type[Analyzer], object]
    options_provider: OptionsProvider

    def __init__(self, stages: Sequence[PipelineStage], options: list):
        self.stages = stages
        self.analyzers_results = {}
        self.options_provider = OptionsProvider()
        self._analyzers = list(
            itertools.chain.from_iterable([stage.analyzers() for stage in stages])
        )
        for option in options:
            self.options_provider.add(option)

    def get_analyzers(self) -> List[Type[Analyzer]]:
        return self._analyzers

    def execute(
        self,
        reference_data: pandas.DataFrame,
        current_data: Optional[pandas.DataFrame] = None,
        column_mapping: Optional[ColumnMapping] = None,
    ) -> None:
        if column_mapping is None:
            column_mapping = ColumnMapping()

        #  making shallow copy - this copy DOES NOT copy existing data, but contains link to it:
        #  - this copy WILL DISCARD all columns changes or rows changes (adding or removing)
        #  - this copy WILL KEEP all values' changes in existing rows and columns.
        rdata = reference_data.copy()
        cdata = None if current_data is None else current_data.copy()
        for analyzer in self.get_analyzers():
            instance = analyzer()
            instance.options_provider = self.options_provider
            self.analyzers_results[analyzer] = instance.calculate(
                rdata, cdata, column_mapping
            )
        for stage in self.stages:
            stage.options_provider = self.options_provider
            stage.calculate(
                rdata.copy(),
                None if cdata is None else cdata.copy(),
                column_mapping,
                self.analyzers_results,
            )



from datetime import datetime

from evidently.analyzers.classification_performance_analyzer import (
    ClassificationPerformanceAnalyzer,
)
from evidently.analyzers.classification_performance_analyzer import (
    ClassificationPerformanceMetrics,
)
from evidently.model_profile.sections.base_profile_section import ProfileSection


class ClassificationPerformanceProfileSection(ProfileSection):
    def part_id(self) -> str:
        return "classification_performance"

    def __init__(self):
        super().__init__()
        self.analyzers_types = [ClassificationPerformanceAnalyzer]
        self._result = None

    def analyzers(self):
        return self.analyzers_types

    @staticmethod
    def _get_performance_metrics_as_dict(metrics: ClassificationPerformanceMetrics):
        return {
            "accuracy": metrics.accuracy,
            "precision": metrics.precision,
            "recall": metrics.recall,
            "f1": metrics.f1,
            "metrics_matrix": metrics.metrics_matrix,
            "confusion_matrix": {
                "labels": metrics.confusion_matrix.labels,
                "values": metrics.confusion_matrix.values,
            },
        }

    def calculate(
        self, reference_data, current_data, column_mapping, analyzers_results
    ):
        result = ClassificationPerformanceAnalyzer.get_results(analyzers_results)
        result_for_json = result.columns.as_dict()
        result_for_json["metrics"] = {}

        if result.reference_metrics:
            result_for_json["metrics"][
                "reference"
            ] = self._get_performance_metrics_as_dict(result.reference_metrics)

        if result.current_metrics:
            result_for_json["metrics"][
                "current"
            ] = self._get_performance_metrics_as_dict(result.current_metrics)

        self._result = {
            "name": self.part_id(),
            "datetime": str(datetime.now()),
            "data": result_for_json,
        }

    def get_results(self):
        return self._result



import dataclasses
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from plotly import graph_objs as go

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import WidgetSize
from evidently.renderers.html_widgets import plotly_figure
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ColumnValuePlotResults:
    column_name: str
    datetime_column_name: Optional[str]
    current_scatter: pd.DataFrame
    reference_scatter: pd.DataFrame


class ColumnValuePlot(Metric[ColumnValuePlotResults]):
    column_name: str

    def __init__(self, column_name: str):
        self.column_name = column_name

    def calculate(self, data: InputData) -> ColumnValuePlotResults:
        if self.column_name not in data.current_data.columns:
            raise ValueError(f"Column '{self.column_name}' should present in the current dataset")

        if data.reference_data is None:
            raise ValueError("Reference data should be present")

        if self.column_name not in data.reference_data.columns:
            raise ValueError(f"Column '{self.column_name}' should present in the reference dataset")

        dataset_columns = process_columns(data.current_data, data.column_mapping)
        if not (
            self.column_name in dataset_columns.num_feature_names
            or (
                self.column_name == dataset_columns.utility_columns.target
                and (data.column_mapping.task == "regression" or is_numeric_dtype(data.current_data[self.column_name]))
            )
            or (
                isinstance(dataset_columns.utility_columns.prediction, str)
                and self.column_name == dataset_columns.utility_columns.prediction
                and (data.column_mapping.task == "regression" or is_numeric_dtype(data.current_data[self.column_name]))
            )
        ):
            raise ValueError("Expected numerical feature")
        datetime_column_name = dataset_columns.utility_columns.date
        curr_df = data.current_data
        ref_df = data.reference_data
        curr_df = self._make_df_for_plot(curr_df, self.column_name, datetime_column_name)
        ref_df = self._make_df_for_plot(ref_df, self.column_name, datetime_column_name)
        return ColumnValuePlotResults(
            column_name=self.column_name,
            datetime_column_name=datetime_column_name,
            current_scatter=curr_df,
            reference_scatter=ref_df,
        )

    def _make_df_for_plot(self, df, column_name: str, datetime_column_name: Optional[str]):
        result = df.replace([np.inf, -np.inf], np.nan)
        if datetime_column_name is not None:
            result.dropna(axis=0, how="any", inplace=True, subset=[column_name, datetime_column_name])
            return result.sort_values(datetime_column_name)
        result.dropna(axis=0, how="any", inplace=True, subset=[column_name])
        return result.sort_index()


@default_renderer(wrap_type=ColumnValuePlot)
class ColumnValuePlotRenderer(MetricRenderer):
    def render_json(self, obj: ColumnValuePlot) -> dict:
        obj.get_result()
        return {}

    def render_html(self, obj: ColumnValuePlot) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        current_scatter = result.current_scatter
        reference_scatter = result.reference_scatter
        column_name = result.column_name

        mean_ref = reference_scatter[column_name].mean()
        std_ref = reference_scatter[column_name].std()
        y0 = mean_ref - std_ref
        y1 = mean_ref + std_ref

        if result.datetime_column_name is not None:
            curr_x = current_scatter[result.datetime_column_name]
            ref_x = reference_scatter[result.datetime_column_name]
            x_name = "Timestamp"

        else:
            curr_x = current_scatter.index
            ref_x = reference_scatter.index
            x_name = "Index"

        color_options = self.color_options

        fig = go.Figure()

        fig.add_trace(
            go.Scattergl(
                x=curr_x,
                y=current_scatter[column_name],
                mode="markers",
                name="Current",
                marker=dict(size=6, color=color_options.get_current_data_color()),
            )
        )
        fig.add_trace(
            go.Scattergl(
                x=ref_x,
                y=reference_scatter[column_name],
                mode="markers",
                name="Reference",
                marker=dict(size=6, color=color_options.get_reference_data_color()),
            )
        )

        x0 = np.max(curr_x)

        fig.add_trace(
            go.Scattergl(
                x=[x0, x0],
                y=[y0, y1],
                mode="markers",
                name="Current",
                marker=dict(size=0.01, color=color_options.non_visible_color, opacity=0.005),
                showlegend=False,
            )
        )

        fig.update_layout(
            xaxis_title=x_name,
            yaxis_title=f"{column_name} value",
            showlegend=True,
            legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
            shapes=[
                dict(
                    type="rect",
                    # x-reference is assigned to the x-values
                    xref="paper",
                    # y-reference is assigned to the plot paper [0,1]
                    yref="y",
                    x0=0,
                    y0=y0,
                    x1=1,
                    y1=y1,
                    fillcolor=color_options.fill_color,
                    opacity=0.5,
                    layer="below",
                    line_width=0,
                ),
                dict(
                    type="line",
                    name="Reference",
                    xref="paper",
                    yref="y",
                    x0=0,  # min(testset_agg_by_date.index),
                    y0=(y0 + y1) / 2,
                    x1=1,  # max(testset_agg_by_date.index),
                    y1=(y0 + y1) / 2,
                    line=dict(color=color_options.zero_line_color, width=3),
                ),
            ],
        )
        return [plotly_figure(title=f"Column '{column_name}' Values", figure=fig, size=WidgetSize.FULL)]



#!/usr/bin/env python
# coding: utf-8
import json
from typing import Optional

import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.cat_target_drift_analyzer import CatTargetDriftAnalyzer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class CatOutputDriftWidget(Widget):
    def __init__(self, title: str, kind: str = "target"):
        super().__init__(title)
        # check kind values and make private
        self.kind = kind  # target or prediction

    def analyzers(self):
        return [CatTargetDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = CatTargetDriftAnalyzer.get_results(analyzers_results)

        if current_data is None:
            raise ValueError("current_data should be present")

        if self.kind == "target":
            if results.columns.utility_columns.target is None:
                result_metrics = None

            else:
                result_metrics = results.target_metrics

        elif self.kind == "prediction":
            if results.columns.utility_columns.prediction is None:
                result_metrics = None

            else:
                result_metrics = results.prediction_metrics

        else:
            raise ValueError('kind should be "target" or "prediction" value')

        if not result_metrics:
            return None

        output_name = result_metrics.column_name
        stattest_name = result_metrics.stattest_name
        drift_score = result_metrics.drift_score
        drift_detected = result_metrics.drift_detected
        output_sim_test = "detected" if drift_detected else "not detected"
        # plot output distributions
        fig = go.Figure()

        fig.add_trace(
            go.Histogram(
                x=reference_data[output_name],
                marker_color=color_options.get_reference_data_color(),
                opacity=0.6,
                nbinsx=10,
                name="Reference",
                histnorm="probability",
            )
        )

        fig.add_trace(
            go.Histogram(
                x=current_data[output_name],
                marker_color=color_options.get_current_data_color(),
                opacity=0.6,
                nbinsx=10,
                name="Current",
                histnorm="probability",
            )
        )

        fig.update_layout(
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
            xaxis_title=output_name.title(),
            yaxis_title="Share",
        )

        output_drift_json = json.loads(fig.to_json())
        return BaseWidgetInfo(
            title=f"{self.kind.title()} Drift: {output_sim_test}, drift score={round(drift_score, 6)} ({stattest_name})",
            type="big_graph",
            size=2,
            params={
                "data": output_drift_json["data"],
                "layout": output_drift_json["layout"],
            },
        )



#!/usr/bin/env python
# coding: utf-8

from typing import Optional

import pandas as pd
import plotly.figure_factory as ff

from evidently import ColumnMapping
from evidently.analyzers.num_target_drift_analyzer import NumTargetDriftAnalyzer
from evidently.dashboard.widgets.utils import CutQuantileTransformer
from evidently.dashboard.widgets.utils import fig_to_json
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions
from evidently.options import QualityMetricsOptions


class NumOutputDriftWidget(Widget):
    def __init__(self, title: str, kind: str = "target"):
        super().__init__(title)
        self.kind = kind  # target or prediction

    def analyzers(self):
        return [NumTargetDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = NumTargetDriftAnalyzer.get_results(analyzers_results)
        quality_metrics_options = self.options_provider.get(QualityMetricsOptions)
        cut_quantile = quality_metrics_options.cut_quantile

        if current_data is None:
            raise ValueError("current_data should be present")

        if self.kind == "target":
            if results.columns.utility_columns.target is None:
                return None

            column_name = results.columns.utility_columns.target
            metrics = results.target_metrics

        elif self.kind == "prediction":
            if results.columns.utility_columns.prediction is None:
                return None

            if not isinstance(results.columns.utility_columns.prediction, str):
                raise ValueError(
                    f"Widget [{self.title}] requires one str value for 'prediction' column"
                )

            column_name = results.columns.utility_columns.prediction
            metrics = results.prediction_metrics

        else:
            raise ValueError(
                f"Widget [{self.title}] requires 'target' or 'prediction' kind parameter value"
            )

        if metrics is None:
            return None

        # calculate output drift
        stattest_name = metrics.stattest_name
        drift_score = metrics.drift_score
        drift_detected = metrics.drift_detected
        output_sim_test = "detected" if drift_detected else "not detected"

        # plot output distributions
        quantile = quality_metrics_options.get_cut_quantile(column_name)
        if cut_quantile and quantile is not None:
            side, q = quantile
            cqt = CutQuantileTransformer(side=side, q=q)
            cqt.fit(reference_data[column_name])
            reference_data_to_plot = cqt.transform(reference_data[column_name])
            current_data_to_plot = cqt.transform(current_data[column_name])
        else:
            reference_data_to_plot = reference_data[column_name]
            current_data_to_plot = current_data[column_name]

        output_distr = ff.create_distplot(
            [reference_data_to_plot, current_data_to_plot],
            ["Reference", "Current"],
            colors=[
                color_options.get_reference_data_color(),
                color_options.get_current_data_color(),
            ],
            show_rug=True,
        )

        output_distr.update_layout(
            xaxis_title="Value",
            yaxis_title="Share",
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
        )

        # output_drift_json = json.loads(output_distr.to_json())
        output_drift_json = fig_to_json(output_distr)

        return BaseWidgetInfo(
            title=f"{self.kind.title()} Drift: {output_sim_test},"
            f" drift score={round(drift_score, 6)} ({stattest_name})",
            type="big_graph",
            details="",
            alerts=[],
            alertsPosition="row",
            insights=[],
            size=2,
            params={
                "data": output_drift_json["data"],
                "layout": output_drift_json["layout"],
            },
            additionalGraphs=[],
        )



from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.metric_preset.metric_preset import MetricPreset
from evidently.metrics import RegressionAbsPercentageErrorPlot
from evidently.metrics import RegressionErrorBiasTable
from evidently.metrics import RegressionErrorDistribution
from evidently.metrics import RegressionErrorNormality
from evidently.metrics import RegressionErrorPlot
from evidently.metrics import RegressionPredictedVsActualPlot
from evidently.metrics import RegressionPredictedVsActualScatter
from evidently.metrics import RegressionQualityMetric
from evidently.metrics import RegressionTopErrorMetric
from evidently.utils.data_operations import DatasetColumns


class RegressionPreset(MetricPreset):
    """Metric preset for Regression performance analysis.

    Contains metrics:
    - RegressionQualityMetric
    - RegressionPredictedVsActualScatter
    - RegressionPredictedVsActualPlot
    - RegressionErrorPlot
    - RegressionAbsPercentageErrorPlot
    - RegressionErrorDistribution
    - RegressionErrorNormality
    - RegressionTopErrorMetric
    - RegressionErrorBiasTable
    """

    columns: Optional[List[str]]

    def __init__(self, columns: Optional[List[str]] = None):
        super().__init__()
        self.columns = columns

    def generate_metrics(self, data: InputData, columns: DatasetColumns):
        return [
            RegressionQualityMetric(),
            RegressionPredictedVsActualScatter(),
            RegressionPredictedVsActualPlot(),
            RegressionErrorPlot(),
            RegressionAbsPercentageErrorPlot(),
            RegressionErrorDistribution(),
            RegressionErrorNormality(),
            RegressionTopErrorMetric(),
            RegressionErrorBiasTable(columns=self.columns),
        ]



import dataclasses
import random
from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import pandas as pd


@dataclasses.dataclass
class SamplingOptions:
    type: str = "none"
    random_seed: int = 1
    ratio: float = 1.0
    n: int = 1


@dataclasses.dataclass
class DataOptions:
    date_column: str
    separator: str
    # is csv file contains header row
    header: bool
    # should be list of names, or None if columns should be inferred from data
    column_names: Optional[List[str]]

    def __init__(self, date_column: str = "datetime", separator=",", header=True, column_names=None):
        self.date_column = date_column
        self.header = header
        self.separator = separator
        self.column_names = column_names


def _skiprows(sampling_options: SamplingOptions) -> Union[Callable[[int], bool], None]:
    if sampling_options.type == "none":
        return None
    if sampling_options.type == "nth":
        if sampling_options.n < 1:
            raise Exception("nth sampling should have 'n' parameter >= 1")
        return __simple(sampling_options)
    if sampling_options.type == "random":
        skip_rows = RandomizedSkipRows(sampling_options.ratio, sampling_options.random_seed)
        return skip_rows.skiprows
    raise ValueError(f"Unexpected sampling type {sampling_options.type}")


def __simple(sampling_options: SamplingOptions):
    def func(row_idx):
        if row_idx == 0:
            result = False
        else:
            rem = row_idx % sampling_options.n
            result = rem != 1
        return result

    return func


class DataLoader:
    def __init__(self):
        pass

    def load(self, filename: str, data_options: DataOptions, sampling_options: SamplingOptions = None):
        sampling_opts = SamplingOptions("none", 0, 0) if sampling_options is None else sampling_options
        parse_dates = [data_options.date_column] if data_options.date_column else False
        return pd.read_csv(
            filename,
            header=0 if data_options.header else None,
            sep=data_options.separator,
            skiprows=_skiprows(sampling_opts),
            parse_dates=parse_dates,
        )


CHUNK_SIZE = 1000


class RandomizedSkipRows:
    def __init__(self, ratio: float, random_seed: int):
        self.random = random.Random(random_seed)
        self.ratio = ratio
        self.selected_rows = self._select()

    def skiprows(self, row_index: int):
        if row_index == 0:
            return False
        if row_index % CHUNK_SIZE == 0:
            self.selected_rows = self._select()
        idx = row_index - int(row_index / CHUNK_SIZE) * CHUNK_SIZE
        return self.selected_rows[idx]

    def _select(self):
        return [bool(self.random.random() < self.ratio) for _ in range(1000)]



#!/usr/bin/env python
# coding: utf-8

from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.reg_abs_perc_error_in_time_widget import (
    RegAbsPercErrorTimeWidget,
)
from evidently.dashboard.widgets.reg_colored_pred_vs_actual_widget import (
    RegColoredPredActualWidget,
)
from evidently.dashboard.widgets.reg_error_distr_widget import RegErrorDistrWidget
from evidently.dashboard.widgets.reg_error_in_time_widget import RegErrorTimeWidget
from evidently.dashboard.widgets.reg_error_normality_widget import (
    RegErrorNormalityWidget,
)
from evidently.dashboard.widgets.reg_pred_and_actual_in_time_widget import (
    RegPredActualTimeWidget,
)
from evidently.dashboard.widgets.reg_pred_vs_actual_widget import RegPredActualWidget
from evidently.dashboard.widgets.reg_quality_metrics_bar_widget import (
    RegQualityMetricsBarWidget,
)
from evidently.dashboard.widgets.reg_underperform_metrics_widget import (
    RegUnderperformMetricsWidget,
)
from evidently.dashboard.widgets.reg_underperform_segments_table_widget import (
    UnderperformSegmTableWidget,
)
from evidently.dashboard.widgets.target_name_widget import TargetNameWidget


class RegressionPerformanceTab(Tab):
    widgets = [
        (
            TargetNameWidget("Regression Model Performance Report.", kind="regression"),
            Verbose.ALWAYS,
        ),
        (
            RegQualityMetricsBarWidget("Reference: Model Quality (+/- std)"),
            Verbose.ALWAYS,
        ),
        (
            RegQualityMetricsBarWidget(
                "Current: Model Quality (+/- std)", dataset="current"
            ),
            Verbose.ALWAYS,
        ),
        (RegPredActualWidget("Reference: Predicted vs Actual"), Verbose.ALWAYS),
        (
            RegPredActualWidget("Current: Predicted vs Actual", dataset="current"),
            Verbose.ALWAYS,
        ),
        (
            RegPredActualTimeWidget("Reference: Predicted vs Actual in Time"),
            Verbose.ALWAYS,
        ),
        (
            RegPredActualTimeWidget(
                "Current: Predicted vs Actual in Time", dataset="current"
            ),
            Verbose.ALWAYS,
        ),
        (RegErrorTimeWidget("Reference: Error (Predicted - Actual)"), Verbose.ALWAYS),
        (
            RegErrorTimeWidget(
                "Current: Error (Predicted - Actual)", dataset="current"
            ),
            Verbose.ALWAYS,
        ),
        (
            RegAbsPercErrorTimeWidget("Reference: Absolute Percentage Error"),
            Verbose.FULL,
        ),
        (
            RegAbsPercErrorTimeWidget(
                "Current: Absolute Percentage Error", dataset="current"
            ),
            Verbose.FULL,
        ),
        (RegErrorDistrWidget("Reference: Error Distribution"), Verbose.ALWAYS),
        (
            RegErrorDistrWidget("Current: Error Distribution", dataset="current"),
            Verbose.ALWAYS,
        ),
        (RegErrorNormalityWidget("Reference: Error Normality"), Verbose.FULL),
        (
            RegErrorNormalityWidget("Current: Error Normality", dataset="current"),
            Verbose.FULL,
        ),
        (
            RegUnderperformMetricsWidget("Reference: Mean Error per Group (+/- std)"),
            Verbose.FULL,
        ),
        (
            RegUnderperformMetricsWidget(
                "Current: Mean Error per Group (+/- std)", dataset="current"
            ),
            Verbose.FULL,
        ),
        (
            RegColoredPredActualWidget("Reference: Predicted vs Actual per Group"),
            Verbose.FULL,
        ),
        (
            RegColoredPredActualWidget(
                "Current: Predicted vs Actual per Group", dataset="current"
            ),
            Verbose.FULL,
        ),
        (
            UnderperformSegmTableWidget(
                "Error Bias: Mean/Most Common Feature Value per Group"
            ),
            Verbose.FULL,
        ),
    ]



import dataclasses
from typing import Dict
from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_drift import get_one_column_drift
from evidently.calculations.stattests import PossibleStatTestType
from evidently.model.widget import BaseWidgetInfo
from evidently.options import DataDriftOptions
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs
from evidently.renderers.render_utils import get_distribution_plot_figure
from evidently.utils.data_operations import process_columns
from evidently.utils.types import Numeric
from evidently.utils.visualizations import Distribution
from evidently.utils.visualizations import plot_scatter_for_data_drift


@dataclasses.dataclass
class ColumnDriftMetricResults:
    column_name: str
    column_type: str
    stattest_name: str
    stattest_threshold: float
    drift_score: Numeric
    drift_detected: bool
    current_distribution: Optional[Distribution]
    reference_distribution: Optional[Distribution]
    current_scatter: Optional[Dict[str, list]]
    x_name: Optional[str]
    plot_shape: Optional[Dict[str, float]]
    typical_examples_cur: Optional[List[str]]
    typical_examples_ref: Optional[List[str]]
    typical_words_cur: Optional[List[str]]
    typical_words_ref: Optional[List[str]]


class ColumnDriftMetric(Metric[ColumnDriftMetricResults]):
    """Calculate drift metric for a column"""

    column_name: str
    stattest: Optional[PossibleStatTestType]
    stattest_threshold: Optional[float]

    def __init__(
        self,
        column_name: str,
        stattest: Optional[PossibleStatTestType] = None,
        stattest_threshold: Optional[float] = None,
    ):
        self.column_name = column_name
        self.stattest = stattest
        self.stattest_threshold = stattest_threshold

    def get_parameters(self) -> tuple:
        return self.column_name, self.stattest_threshold, self.stattest

    def calculate(self, data: InputData) -> ColumnDriftMetricResults:
        if data.reference_data is None:
            raise ValueError("Reference dataset should be present")

        if self.column_name not in data.current_data:
            raise ValueError(f"Cannot find column '{self.column_name}' in current dataset")

        if self.column_name not in data.reference_data:
            raise ValueError(f"Cannot find column '{self.column_name}' in reference dataset")

        dataset_columns = process_columns(data.reference_data, data.column_mapping)
        options = DataDriftOptions(all_features_stattest=self.stattest, threshold=self.stattest_threshold)
        drift_result = get_one_column_drift(
            current_data=data.current_data,
            reference_data=data.reference_data,
            column_name=self.column_name,
            dataset_columns=dataset_columns,
            options=options,
        )

        return ColumnDriftMetricResults(
            column_name=drift_result.column_name,
            column_type=drift_result.column_type,
            stattest_name=drift_result.stattest_name,
            stattest_threshold=drift_result.threshold,
            drift_score=drift_result.drift_score,
            drift_detected=drift_result.drift_detected,
            current_distribution=drift_result.current_distribution,
            reference_distribution=drift_result.reference_distribution,
            current_scatter=drift_result.current_scatter,
            x_name=drift_result.x_name,
            plot_shape=drift_result.plot_shape,
            typical_examples_cur=drift_result.typical_examples_cur,
            typical_examples_ref=drift_result.typical_examples_ref,
            typical_words_cur=drift_result.typical_words_cur,
            typical_words_ref=drift_result.typical_words_ref,
        )


@default_renderer(wrap_type=ColumnDriftMetric)
class ColumnDriftMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnDriftMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        # remove distribution data with pandas dataframes
        result.pop("current_distribution", None)
        result.pop("reference_distribution", None)
        result.pop("current_scatter", None)
        result.pop("x_name", None)
        result.pop("plot_shape", None)
        result.pop("typical_examples_cur", None)
        result.pop("typical_examples_ref", None)
        result.pop("typical_words_cur", None)
        result.pop("typical_words_ref", None)
        return result

    def render_html(self, obj: ColumnDriftMetric) -> List[BaseWidgetInfo]:
        result = obj.get_result()

        if result.drift_detected:
            drift = "detected"

        else:
            drift = "not detected"

        drift_score = round(result.drift_score, 3)

        tabs = []

        # fig_json = fig.to_plotly_json()
        if result.current_scatter is not None and result.plot_shape is not None and result.x_name is not None:
            scatter_fig = plot_scatter_for_data_drift(
                curr_y=result.current_scatter[result.column_name],
                curr_x=result.current_scatter[result.x_name],
                y0=result.plot_shape["y0"],
                y1=result.plot_shape["y1"],
                y_name=result.column_name,
                x_name=result.x_name,
                color_options=self.color_options,
            )
            tabs.append(TabData("DATA DRIFT", plotly_figure(title="", figure=scatter_fig)))

        if result.current_distribution is not None and result.reference_distribution is not None:
            distr_fig = get_distribution_plot_figure(
                current_distribution=result.current_distribution,
                reference_distribution=result.reference_distribution,
                color_options=self.color_options,
            )
            # figures.append(GraphData.figure("DATA DISTRIBUTION", distr_fig))
            tabs.append(TabData("DATA DISTRIBUTION", plotly_figure(title="", figure=distr_fig)))

        if (
            result.typical_examples_cur is not None
            and result.typical_examples_ref is not None
            and result.typical_words_cur is not None
            and result.typical_words_ref is not None
        ):
            current_table_words = table_data(
                title="",
                column_names=["", ""],
                data=[[el, ""] for el in result.typical_words_cur],
            )
            reference_table_words = table_data(
                title="",
                column_names=["", ""],
                data=[[el, ""] for el in result.typical_words_ref],
            )
            current_table_examples = table_data(
                title="",
                column_names=["", ""],
                data=[[el, ""] for el in result.typical_examples_cur],
            )
            reference_table_examples = table_data(
                title="",
                column_names=["", ""],
                data=[[el, ""] for el in result.typical_examples_ref],
            )

            tabs = [
                TabData(title="current: characteristic words", widget=current_table_words),
                TabData(title="reference: characteristic words", widget=reference_table_words),
                TabData(title="current: characteristic examples", widget=current_table_examples),
                TabData(title="reference: characteristic examples", widget=reference_table_examples),
            ]
        render_result = [
            counter(
                counters=[
                    CounterData(
                        (
                            f"Data drift {drift}. "
                            f"Drift detection method: {result.stattest_name}. "
                            f"Drift score: {drift_score}"
                        ),
                        f"Drift in column '{result.column_name}'",
                    )
                ],
                title="",
            )
        ]
        if len(tabs) > 0:
            render_result.append(
                widget_tabs(
                    title="",
                    tabs=tabs,
                )
            )

        return render_result



import dataclasses
from typing import Dict
from typing import List
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import ColumnCorrelations
from evidently.calculations.data_quality import calculate_category_column_correlations
from evidently.calculations.data_quality import calculate_numerical_column_correlations
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import get_histogram_for_distribution
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import widget_tabs
from evidently.utils.data_operations import process_columns
from evidently.utils.data_operations import recognize_column_type


@dataclasses.dataclass
class ColumnCorrelationsMetricResult:
    column_name: str
    current: Dict[str, ColumnCorrelations]
    reference: Optional[Dict[str, ColumnCorrelations]] = None


class ColumnCorrelationsMetric(Metric[ColumnCorrelationsMetricResult]):
    """Calculates correlations between the selected column and all the other columns.
    In the current and reference (if presented) datasets"""

    column_name: str

    def __init__(self, column_name: str) -> None:
        self.column_name = column_name

    @staticmethod
    def _calculate_correlation(
        column_name: str, dataset: pd.DataFrame, column_mapping: ColumnMapping
    ) -> Dict[str, ColumnCorrelations]:
        columns = process_columns(dataset, column_mapping)
        column_type = recognize_column_type(dataset=dataset, column_name=column_name, columns=columns)

        if column_type == "cat":
            correlation_columns = [name for name in columns.cat_feature_names if name != column_name]
            return calculate_category_column_correlations(column_name, dataset, correlation_columns)

        elif column_type == "num":
            correlation_columns = [name for name in columns.num_feature_names if name != column_name]
            return calculate_numerical_column_correlations(column_name, dataset, correlation_columns)

        else:
            raise ValueError(f"Cannot calculate correlations for '{column_type}' column type.")

    def calculate(self, data: InputData) -> ColumnCorrelationsMetricResult:
        if self.column_name not in data.current_data:
            raise ValueError(f"Column '{self.column_name}' was not found in current data.")

        if data.reference_data is not None:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column '{self.column_name}' was not found in reference data.")

        current_correlations = self._calculate_correlation(self.column_name, data.current_data, data.column_mapping)

        if data.reference_data is not None:
            reference_correlations: Optional[Dict[str, ColumnCorrelations]] = self._calculate_correlation(
                self.column_name, data.reference_data, data.column_mapping
            )

        else:
            reference_correlations = None

        return ColumnCorrelationsMetricResult(
            column_name=self.column_name,
            current=current_correlations,
            reference=reference_correlations,
        )


@default_renderer(wrap_type=ColumnCorrelationsMetric)
class ColumnCorrelationsMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnCorrelationsMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    def _get_plots_correlations(self, metric_result: ColumnCorrelationsMetricResult) -> Optional[BaseWidgetInfo]:
        tabs = []

        for correlation_name, current_correlation in metric_result.current.items():
            reference_correlation_values = None

            if metric_result.reference and correlation_name in metric_result.reference:
                reference_correlation_values = metric_result.reference[correlation_name].values

            if current_correlation.values or reference_correlation_values:
                tabs.append(
                    TabData(
                        title=correlation_name,
                        widget=get_histogram_for_distribution(
                            title="",
                            current_distribution=current_correlation.values,
                            reference_distribution=reference_correlation_values,
                            xaxis_title="Columns",
                            yaxis_title="Correlation",
                            color_options=self.color_options,
                        ),
                    )
                )

        if tabs:
            return widget_tabs(tabs=tabs)

        else:
            return None

    def render_html(self, obj: ColumnCorrelationsMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        correlation_plots = self._get_plots_correlations(metric_result)

        if correlation_plots:
            return [header_text(label=f"Correlations for column '{metric_result.column_name}'."), correlation_plots]

        else:
            # no correlations, draw nothing
            return []



import dataclasses
from typing import List
from typing import Optional

import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.classification_performance import PredictionData
from evidently.calculations.classification_performance import calculate_pr_table
from evidently.calculations.classification_performance import get_prediction_data
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import WidgetSize
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ClassificationPRTableResults:
    current_pr_table: Optional[dict] = None
    reference_pr_table: Optional[dict] = None


class ClassificationPRTable(Metric[ClassificationPRTableResults]):
    def calculate(self, data: InputData) -> ClassificationPRTableResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        curr_prediction = get_prediction_data(data.current_data, dataset_columns, data.column_mapping.pos_label)
        curr_pr_table = self.calculate_metrics(data.current_data[target_name], curr_prediction)
        ref_pr_table = None
        if data.reference_data is not None:
            ref_prediction = get_prediction_data(data.reference_data, dataset_columns, data.column_mapping.pos_label)
            ref_pr_table = self.calculate_metrics(data.reference_data[target_name], ref_prediction)
        return ClassificationPRTableResults(
            current_pr_table=curr_pr_table,
            reference_pr_table=ref_pr_table,
        )

    def calculate_metrics(self, target_data: pd.Series, prediction: PredictionData):
        labels = prediction.labels
        if prediction.prediction_probas is None:
            raise ValueError("PR Table can be calculated only on binary probabilistic predictions")
        binaraized_target = (target_data.values.reshape(-1, 1) == labels).astype(int)
        pr_table = {}
        if len(labels) <= 2:
            binaraized_target = pd.DataFrame(binaraized_target[:, 0])
            binaraized_target.columns = ["target"]

            binded = list(zip(binaraized_target["target"].tolist(), prediction.prediction_probas.iloc[:, 0].tolist()))
            pr_table[prediction.prediction_probas.columns[0]] = calculate_pr_table(binded)
        else:
            binaraized_target = pd.DataFrame(binaraized_target)
            binaraized_target.columns = labels

            for label in labels:
                binded = list(zip(binaraized_target[label].tolist(), prediction.prediction_probas[label]))
                pr_table[label] = calculate_pr_table(binded)
        return pr_table


@default_renderer(wrap_type=ClassificationPRTable)
class ClassificationPRTableRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationPRTable) -> dict:
        return {}

    def render_html(self, obj: ClassificationPRTable) -> List[BaseWidgetInfo]:
        reference_pr_table = obj.get_result().reference_pr_table
        current_pr_table = obj.get_result().current_pr_table
        columns = ["Top(%)", "Count", "Prob", "TP", "FP", "Precision", "Recall"]
        result = []
        size = WidgetSize.FULL
        # if reference_pr_table is not None:
        #     size = WidgetSize.HALF
        if current_pr_table is not None:
            if len(current_pr_table.keys()) == 1:
                result.append(
                    table_data(
                        column_names=columns,
                        data=current_pr_table[list(current_pr_table.keys())[0]],
                        title="Current: Precision-Recall Table",
                        size=size,
                    )
                )
            else:
                tab_data = []
                for label in current_pr_table.keys():
                    table = table_data(column_names=columns, data=current_pr_table[label], title="", size=size)
                    tab_data.append(TabData(label, table))
                result.append(widget_tabs(title="Current: Precision-Recall Table", tabs=tab_data))
        if reference_pr_table is not None:
            if len(reference_pr_table.keys()) == 1:
                result.append(
                    table_data(
                        column_names=columns,
                        data=reference_pr_table[list(reference_pr_table.keys())[0]],
                        title="Reference: Precision-Recall Table",
                        size=size,
                    )
                )
            else:
                tab_data = []
                for label in reference_pr_table.keys():
                    table = table_data(column_names=columns, data=reference_pr_table[label], title="", size=size)
                    tab_data.append(TabData(label, table))
                result.append(widget_tabs(title="Reference: Precision-Recall Table", tabs=tab_data))
        return result



#!/usr/bin/env python
# coding: utf-8
import abc
from dataclasses import dataclass
from typing import Optional

import pandas as pd

from evidently.options import OptionsProvider
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.utils.data_operations import DatasetColumns


@dataclass
class BaseAnalyzerResult:
    """Base class for all analyzers results.

    If you want to add a new analyzer, inherit a results class from the class.
    For correct initiation you should add a decorator `@dataclass` to children classes too.

        For example:

        @dataclass
        class RegressionPerformanceAnalyzerResults(BaseAnalyzerResult):
            my_result: str
    """

    columns: DatasetColumns


class Analyzer:
    @abc.abstractmethod
    def calculate(
        self, reference_data: pd.DataFrame, current_data: Optional[pd.DataFrame], column_mapping: ColumnMapping
    ) -> BaseAnalyzerResult:
        raise NotImplementedError()

    options_provider: OptionsProvider



import abc
import functools
from typing import Dict
from typing import List
from typing import Tuple
from typing import Type

from evidently.base_metric import Metric
from evidently.tests.base_test import Test


class ExecutionGraph:
    @abc.abstractmethod
    def get_metric_execution_iterator(self) -> List[Tuple[Metric, Metric]]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_test_execution_iterator(self) -> List[Test]:
        raise NotImplementedError()


class SimpleExecutionGraph(ExecutionGraph):
    """
    Simple execution graph without any work with dependencies at all,
     assumes that metrics already in order for execution
    """

    metrics: List[Metric]
    tests: List[Test]

    def __init__(self, metrics: List[Metric], tests: List[Test]):
        self.metrics = metrics
        self.tests = tests

    def get_metric_execution_iterator(self) -> List[Tuple[Metric, Metric]]:
        aggregated: Dict[Type[Metric], List[Metric]] = functools.reduce(_aggregate_metrics, self.metrics, {})
        metric_to_calculations = {}
        for metric_type, metrics in aggregated.items():
            metrics_by_parameters: Dict[tuple, List[Metric]] = functools.reduce(_aggregate_by_parameters, metrics, {})

            for metric in metrics:
                parameters = metric.get_parameters()
                if parameters is None:
                    metric_to_calculations[metric] = metric
                else:
                    metric_to_calculations[metric] = metrics_by_parameters[parameters][0]

        return [(metric, metric_to_calculations[metric]) for metric in self.metrics]

    def get_test_execution_iterator(self) -> List[Test]:
        return self.tests


def _aggregate_metrics(agg, item):
    agg[type(item)] = agg.get(type(item), []) + [item]
    return agg


def _aggregate_by_parameters(agg: dict, metric: Metric) -> dict:
    agg[metric.get_parameters()] = agg.get(metric.get_parameters(), []) + [metric]
    return agg



from typing import Generator

from evidently.analyzers.cat_target_drift_analyzer import CatTargetDriftAnalyzer
from evidently.model_monitoring.monitoring import MetricsType
from evidently.model_monitoring.monitoring import ModelMonitor
from evidently.model_monitoring.monitoring import ModelMonitoringMetric


class CatTargetDriftMonitorMetrics:
    """Class for category target grift metrics.

    Metrics list:
        - count: quantity of rows in `reference` and `current` datasets
        - drift: p_value for the data drift
    """

    _tag = "cat_target_drift"
    count = ModelMonitoringMetric(f"{_tag}:count", ["dataset"])
    drift = ModelMonitoringMetric(f"{_tag}:drift", ["kind"])


class CatTargetDriftMonitor(ModelMonitor):
    def monitor_id(self) -> str:
        return "cat_target_drift"

    def analyzers(self):
        return [CatTargetDriftAnalyzer]

    def metrics(self, analyzer_results) -> Generator[MetricsType, None, None]:
        results = CatTargetDriftAnalyzer.get_results(analyzer_results)

        # quantity of rows in income data
        yield CatTargetDriftMonitorMetrics.count.create(
            results.reference_data_count, dict(dataset="prediction")
        )
        yield CatTargetDriftMonitorMetrics.count.create(
            results.current_data_count, dict(dataset="current")
        )

        if results.prediction_metrics:
            yield CatTargetDriftMonitorMetrics.drift.create(
                results.prediction_metrics.drift_score, dict(kind="prediction")
            )

        if results.target_metrics:
            yield CatTargetDriftMonitorMetrics.drift.create(
                results.target_metrics.drift_score, dict(kind="target")
            )



"""Hellinger distance of two samples.

Name: "hellinger"

Import:

    >>> from evidently.calculations.stattests import hellinger_stat_test

Properties:
- only for categorical and numerical features
- returns distance

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import hellinger_stat_test
    >>> options = DataDriftOptions(all_features_stattest=hellinger_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="hellinger")
"""
from collections import defaultdict
from math import sqrt
from typing import DefaultDict
from typing import Tuple

import numpy as np
import pandas as pd

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _hellinger_distance(
    reference_data: pd.Series,
    current_data: pd.Series,
    feature_type: str,
    threshold: float,
) -> Tuple[float, bool]:
    """Compute the Hellinger distance between two arrays
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
    Returns:
        hellinger_distance: normed Hellinger distance
        test_result: whether the drift is detected
    """
    reference_data.dropna(inplace=True)
    current_data.dropna(inplace=True)

    keys = list((set(reference_data.unique()) | set(current_data.unique())))

    if feature_type == "cat":
        dd: DefaultDict[int, int] = defaultdict(int)
        ref = (reference_data.value_counts() / len(reference_data)).to_dict(dd)
        curr = (current_data.value_counts() / len(current_data)).to_dict(dd)

        hellinger_distance = 0.0
        for key in keys:
            p1 = ref[key]
            p2 = curr[key]
            hellinger_distance += sqrt(p1 * p2)

        hellinger_distance = np.clip(hellinger_distance, 0, 1)
        hellinger_distance = sqrt(1 - hellinger_distance)

    else:
        bins = np.histogram_bin_edges(keys, bins="sturges")
        h1 = np.histogram(reference_data.values, bins=bins, density=True)[0]
        h2 = np.histogram(current_data.values, bins=bins, density=True)[0]

        bin_width = (max(bins) - min(bins)) / (len(bins) - 1)

        hellinger_distance = 0.0
        for i in range(len(h1)):
            p1 = h1[i]
            p2 = h2[i]
            hellinger_distance += sqrt(p1 * p2) * bin_width

        hellinger_distance = np.clip(hellinger_distance, 0, 1)
        hellinger_distance = sqrt(1 - hellinger_distance)

    return hellinger_distance, hellinger_distance >= threshold


hellinger_stat_test = StatTest(
    name="hellinger",
    display_name="Hellinger distance",
    func=_hellinger_distance,
    allowed_feature_types=["cat", "num"],
    default_threshold=0.1,
)

register_stattest(hellinger_stat_test)



#!/usr/bin/env python
# coding: utf-8
from dataclasses import dataclass
from typing import Optional
from typing import Sequence

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.base_analyzer import BaseAnalyzerResult
from evidently.calculations.data_drift import ColumnDataDriftMetrics
from evidently.calculations.data_drift import ensure_prediction_column_is_string
from evidently.calculations.data_drift import get_one_column_drift
from evidently.calculations.data_quality import get_rows_count
from evidently.options import DataDriftOptions
from evidently.options import QualityMetricsOptions
from evidently.utils.data_operations import process_columns
from evidently.utils.data_operations import replace_infinity_values_to_nan


@dataclass
class CatTargetDriftAnalyzerResults(BaseAnalyzerResult):
    """Class for all results of category target drift calculations"""

    target_metrics: Optional[ColumnDataDriftMetrics] = None
    prediction_metrics: Optional[ColumnDataDriftMetrics] = None
    reference_data_count: int = 0
    current_data_count: int = 0


class CatTargetDriftAnalyzer(Analyzer):
    """Categorical target drift analyzer.

    Analyze categorical `target` and `prediction` distributions and provide calculations to the following questions:
    Does the model target behave similarly to the past period? Do my model predictions still look the same?

    For reference see https://evidentlyai.com/blog/evidently-014-target-and-prediction-drift
    """

    @staticmethod
    def get_results(analyzer_results) -> CatTargetDriftAnalyzerResults:
        return analyzer_results[CatTargetDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
    ) -> CatTargetDriftAnalyzerResults:
        """Calculate the target and prediction drifts.

        With default options, uses a chi² test when number of labels is greater than 2.
        Otherwise, uses a z-test.

        Notes:
            Be aware that any nan or infinity values will be dropped from the dataframes in place.

            You can also provide a custom function that computes a statistic by adding special
            `DataDriftOptions` object to the `option_provider` of the class.::

                options = DataDriftOptions(cat_target_stattest_func=...)
                analyzer.options_prover.add(options)

            Such a function takes two arguments:::

                def(reference_data: pd.Series, current_data: pd.Series):
                   ...

            and returns arbitrary number (like a p_value from the other tests ;-))
        Args:
            reference_data: usually the data which you used in training.
            current_data: new, unseen data to which we compare the reference data.
            column_mapping: a `ColumnMapping` object that contains references to the name of target and prediction
                columns
        Returns:
            A dictionary that contains some meta information as well as `metrics` for either target or prediction
            columns or both. The `*_drift` column in `metrics` contains a computed p_value from tests.
        """
        if reference_data is None:
            raise ValueError("reference_data should be present")

        if current_data is None:
            raise ValueError("current_data should be present")

        data_drift_options = self.options_provider.get(DataDriftOptions)
        columns = process_columns(reference_data, column_mapping)
        target_column = columns.utility_columns.target

        if not isinstance(target_column, str) and isinstance(target_column, Sequence):
            raise ValueError("target should not be a sequence")

        classification_threshold = self.options_provider.get(QualityMetricsOptions).classification_threshold
        prediction_column = ensure_prediction_column_is_string(
            prediction_column=columns.utility_columns.prediction,
            current_data=current_data,
            reference_data=reference_data,
            threshold=classification_threshold,
        )

        result = CatTargetDriftAnalyzerResults(
            columns=columns,
            reference_data_count=get_rows_count(reference_data),
            current_data_count=get_rows_count(current_data),
        )

        # consider replacing only values in target and prediction column
        reference_data = replace_infinity_values_to_nan(reference_data)
        current_data = replace_infinity_values_to_nan(current_data)

        if target_column is not None:
            result.target_metrics = get_one_column_drift(
                current_data=current_data,
                reference_data=reference_data,
                column_name=target_column,
                dataset_columns=columns,
                options=data_drift_options,
                column_type="cat",
            )

        if prediction_column is not None:
            result.prediction_metrics = get_one_column_drift(
                current_data=current_data,
                reference_data=reference_data,
                column_name=prediction_column,
                dataset_columns=columns,
                options=data_drift_options,
                column_type="cat",
            )

        return result
